"""Given this has been a reoccuring trend where the confusion matrix needs
calculated and certain metrics need calculated FROM the confusion matrix, add
__tested__ metrics derived from the confusion matrix for efficient
computation.
"""
from copy import copy
import os

import h5py
import numpy as np
import pandas as pd
from plotly import graph_objects as go
from scipy.stats import gmean, entropy
from sklearn.metrics import confusion_matrix

from exputils.io import create_filepath
from exputils.data.labels import NominalDataEncoder as NDE

import logging
logger = logging.getLogger(__name__)

# TODO ConfusionTensor: generalize confusion matrix to multiple discrete RVs
# with a ConfusionTensor class, which would be the parent to ConfusionMatrix

# TODO consider ways to interlink this with LabelEncoder better. If necessary


class ConfusionMatrix(object):
    """Confusion matrix of occurrences for nominal data.
    Rows are the known labels and columns are the predictions.
    """
    def __init__(self, targets, preds=None, labels=None, *args, **kwargs):
        """
        Parameters
        ----------
        targets : np.ndarray()
            The confusion matrix to wrap, or the target labels.
        preds : np.ndarray, optional
        labels : np.ndarray, optional
            The labels for the row and columns of the confusion matrix
        """

        # TODO Add optional class weights attribute, allowing it to be
        # overwritten for when the measures are claculated via a param

        # TODO add sample weights param to init here to affect calc of conf

        # TODO perhaps add a matrix of belief/reliability per pairwise class
        # expressed by some probability per element

        if preds is None:
            if (
                (
                    isinstance(targets, np.ndarray)
                    or isinstance(targets, pd.DataFrame)
                )
                and len(targets.shape) == 2
                and targets.shape[0] == targets.shape[1]
            ):
                # If given an existing matrix as a confusion matrix
                self.mat = np.array(targets)

                if isinstance(targets, pd.DataFrame):
                    self.labels = NDE(targets.columns)
                elif isinstance(labels, (list, np.ndarray)):
                    self.labels = np.array(labels)
                    #self.labels = NDE(labels)
                else:
                    self.labels = NDE(np.arange(len(self.mat)))
            else:
                raise TypeError(' '.join([
                    'targets type is expected to be of type `np.ndarray`, but',
                    f'recieved type of {type(targets)}.',
                ]))
        elif preds is not None:
            # Calculate the confusion matrix from targets and preds with sklearn
            if isinstance(labels, (list, np.ndarray)):
                self.labels = np.array(labels)
                #self.labels = NDE(labels)
            else:
                self.labels = np.array(list(set(targets) | set(preds)))
                #self.labels = NDE(list(set(targets) | set(preds)))

            self.mat = confusion_matrix(
                targets,
                preds,
                labels=np.array(self.labels),
                *args,
                **kwargs,
            )

    def __add__(self, other):
        """Add two ConfusionMatrices together if of the same shape w/ same
        label set."""
        if not isinstance(other, ConfusionMatrix):
            raise TypeError(
                'Operator add only supported between two ConfusionMatrices.'
            )

        if len(self.labels) != len(other.labels):
            raise ValueError(' '.join([
                'The two ConfusionMatrices do not have the same number of',
                'labels!',
            ]))
        if all(self.labels != other.labels):
            if set(self.labels) != set(other.labels):
                raise ValueError(
                    'The other ConfusionMatrix does not have the same labels!'
                )

            # TODO reorganize other to be added correctly to existing conf mat
            raise NotImplementedError('Same labels of different order.')

        return ConfusionMatrix(
            self.mat + other.mat,
            labels=copy(self.labels),
        )

    def reduce(self, labels, reduced_label, inverse=False):
        """Reduce confusion matrix to smaller size by mapping labels to one.

        Parameters
        ----------
        labels : list | np.ndarray
            Labels to be reduced to a single label. If `inverse` is True, then
            all the labels in the confusion matrix NOT included in this list
            are to be reduced.
        reduced_label : object
            The label to replace the others.
        inverse : bool, optional
            If True, all of the labels in the confusion matrix NOT in `labels`
            are reduced instead.

        Returns
        -------
        ConfusionMatrix
            The resulting reduced confusion matrix.
        """
        if inverse:
            if reduced_label in self.labels and reduced_label in labels:
                labels = np.delete(
                    labels,
                    np.where(labels == reduced_label)[0][0],
                )
        else:
            if reduced_label in self.labels and reduced_label not in labels:
                labels = np.append(labels, reduced_label)

        # TODO the use of the nominal label encoder would be good here.
        mask = np.zeros(len(self.labels)).astype(bool)
        for label in labels:
            mask |= self.labels == label

        if inverse:
            # Numpy mask in where= ignores False from the calculation.
            not_mask = mask
            mask = np.logical_not(mask)
        else:
            not_mask = np.logical_not(mask)

        # Construct reduced st reduced label is last index
        reduced_cm = np.vstack((
            np.hstack((
                self.mat[not_mask][:, not_mask],
                self.mat[not_mask][:, mask].sum(1, keepdims=True),
            )),
            np.hstack((
                self.mat[mask][:, not_mask].sum(0, keepdims=True),
                self.mat[mask][:, mask].sum(keepdims=True),
            )),
        ))

        return ConfusionMatrix(
            reduced_cm,
            labels=np.append(self.labels[not_mask], reduced_label),
        )

    # TODO methods for the metrics able to be derived from the confusion matrix
    # f score (1 and beta)
    # informedness # This may not generalize to multi class
    # ROC
    # ROC AUC
    # TOC
    # TOC AUC

    def accuracy(self, label_weights=None):
        #return (self.true_pos + self.true_negatives) / all = Trues / all
        if label_weights is not None:
            raise NotImplementedError('Use sklearn.metrics on the samples')
        return np.diagonal(self.mat).sum() / self.mat.sum()

    def error_rate(self, label_weights=None):
        if label_weights is not None:
            raise NotImplementedError('Use sklearn.metrics on the samples')
        return 1.0 - self.accuracy()

    def true_rate(self, label=None, average=False, label_weights=None):
        """Recall, sensitivity, hit rate, or true positive rate. This is
        calculated the same as specificity, selectivity or true negative rate,
        but on the different classes.

        Parameters
        ----------
        average : bool
            if True, averages all true class rates together. Otherwise, returns
            the true rate per class in order of labels.
        """
        # TODO check this and make a unit test for it.
        #recalls = np.diagonal(self.mat) / self.mat.sum(axis=1)
        if label is None:
            tp = self.true_positive()
            return tp / (tp + self.false_negative())

        tp = self.true_positive(label)
        return tp / (tp + self.false_negative(label))

        """
        if average:
            if label_weights is not None:
                raise NotImplementedError('Use sklearn.metrics on the samples')
            # Provide the averaged true class rates (balanced accuracy)
            return recalls.mean()

        # Provide the recall per class
        return recalls
        """

    def true_positive(self, label=None):
        if label is None:
            return np.diag(self.mat)
        return np.diag(self.mat)[label]

    def false_positive(self, label=None):
        """Calculates the False Positives either of every class or a specified
        slice.

        Parameters
        ----------
        label : np.ndarry slice
            A slice of the conufsion matrix indices, NOT the labels themselves
        """
        # TODO again, support for NominalDataEncoder would make labels able to
        # support the actual labels
        if label is None:
            return self.mat.sum(0) - np.diag(self.mat)
        return self.mat[label].sum() - self.mat[label, label]

    def false_negative(self, label=None):
        """Calculates the False Negatives of every class or a specified slice.
        """
        if label is None:
            return self.mat.sum(1) - np.diag(self.mat)
        mask = np.ones(self.mat.shape[0]).astype(bool)
        mask[label] = False
        return self.mat[:, label].sum() - self.mat[label, label]

    # NOTE True Negatives for one class has overlap between the other classes'
    # counts as per class it becomes the sum of everything outside that
    # classes' row and column index, naturally resulting in overlap in counts
    # to other classes true negatives.

    def false_rates(self, label):
        """Calcuate the False Positive/Negative Rates for a single label."""
        tp = self.true_positive(label)
        fp = self.false_positive(label)
        fn = self.false_negative(label)
        tn = self.mat.sum() - tp - fp - fn

        logger.debug('tp = %f', tp)
        logger.debug('tn = %f', tn)
        logger.debug('fp = %f', fp)
        logger.debug('fn = %f', fn)

        fpr = fp / (fp + tn)
        fnr = fn / (tp + fn)

        return fpr, fnr

    def f_score(self, beta=1, average='macro', label_weights=None):
        # average match sklearn f1_score: None, "binary", micro, macro,
        # weighted
        raise NotImplementedError('Use sklearn.metrics on the samples')

    def mcc(self, label_weights=None):
        """Mathew's Correlation Coefficient generalized to multiple classes,
        R_k, a generalizatin of Pearson's correlation coefficient.
        """
        if label_weights is not None:
            raise NotImplementedError()

        actual = self.mat.sum(1)
        predicted = self.mat.sum(0)
        correct_pred = np.diagonal(self.mat).sum()
        total_sqrd = self.mat.sum()**2

        # TODO be aware that NaNs occur at times, and the below may need a
        # nonzero mask like mutual information did!
        denominator = (
            np.sqrt(total_sqrd - np.dot(predicted, predicted))
            * np.sqrt(total_sqrd - np.dot(actual, actual))
        )
        if denominator == 0:
            return 0
        return (
            (correct_pred * self.mat.sum() - np.dot(actual, predicted)) /
            denominator
        )

    def mutual_information(self, normalized=None, weights=None, base=None):
        """The confusion matrix is the joint probability mass function when
        the values are divided by the total of the confusion matrix.

        Parameters
        ----------
        normalized : str, optional
            The method of normalizing the mutual information. Options include:
            'arithmetic' mean, 'geometric' mean, 'min'imum entropy, 'max'imum
            entropy, 'add' the entropies to obtain the redundancy, 'harmonic'
            mean of the two uncertainty coefficients, 'information quality
            ratio' aka a normalized mutual information as a special case of
            dual total correlation.
        weights : np.ndarray, optional
            A 2 by 2 array to weight each element of the discrete joint
            distribution.
        base : {None, 2}
            Whether to use the natural base or base 2. Defaults to 'e' for the
            natural logarithm to match with default expectation of
            scikit-learn.metrics.mutual_info_score.
        """
        # TODO weighted variants, more normalized variants, adjusted, etc.
        # TODO replicate the normalization method in scikitlearn but from a
        # given confusion matrix.
        #   NMI using "arithemtic" in scikit learn is: MI / mean(H(X),H(Y))
        #   thus, it follow geometric mean is similarly used, as is min() and
        #   max()

        # TODO I did the check and they are close but slight difference occurs
        # at the 16th decimal place and on, which make sense cuz of float
        # errors, but I am uncertain if they have a workaround to ensure it is
        # more exact in the scikit-learn code.
        # Using test example actual = [3, 2, 0, 0, 4] and pred = [4, 2, 3, 0,
        # 1], The below MI and NMI for min, max, & arithmetic all matched ==,
        # but geometric did not match exactly. it was off by ~1e-16.
        # This may be due to not removing zero prob items from  marginals in
        # entropy calc?
        # Oh hey, the above is one unit test I can add. LOL

        if base is None:
            log = np.log
        elif base == 2:
            log = np.log2
        else:
            raise ValueError(f'Unexpected value for `base`: {base}')

        joint_distrib = self.mat / self.mat.sum()
        marginal_actual = joint_distrib.sum(1).reshape(-1,1)
        marginal_pred = joint_distrib.sum(0).reshape(1,-1)

        # TODO need a unit test for this; have confirmed matches
        # sklearn.metrics.mutual_info_score, tho, when using natural base 'e'
        denom = np.dot(marginal_actual, marginal_pred)

        # Mask nonzero joint to avoid unnecessary zero division
        nonzero = joint_distrib != 0

        if (
            isinstance(weights, np.ndarray)
            and weights.shape == joint_distrib.shape
        ):
            # Weighted MI from Guiasu 1977
            mutual_info = (
                weights.flatten() * joint_distrib[nonzero]
                * log(joint_distrib[nonzero] / denom[nonzero])
            ).sum()
        else:
            mutual_info = (
                joint_distrib[nonzero]
                * log(joint_distrib[nonzero] / denom[nonzero])
            ).sum()

        # TODO able to avoid re-calc of marginals by simply calling scipy
        # entropy here! Also avoiding two func calls.
        if normalized is None:
            return mutual_info
        if normalized == 'arithmetic':
            return mutual_info / np.mean(
                (self.entropy(0, base=base), self.entropy(1, base=base))
            )
        if normalized == 'geometric':
            return mutual_info / gmean(
                (self.entropy(0, base=base), self.entropy(1, base=base))
            )
        if normalized == 'min':
            return mutual_info / np.minimum(
                self.entropy(0, base=base),
                self.entropy(1, base=base),
            )
        if normalized == 'max':
            return mutual_info / np.maximum(
                self.entropy(0, base=base),
                self.entropy(1, base=base),
            )
        if normalized == 'add':
            # Redundancy
            return mutual_info / (
                self.entropy(0, base=base) + self.entropy(1, base=base)
            )
        if normalized == 'harmonic':
            # Symmetric uncertainty: harmonic mean of the 2 uncertainty coef.s
            return mutual_info / (2 * (
                self.entropy(0, base=base) + self.entropy(1, base=base)
            ))
        if normalized == 'information quality ratio':
            # Information Quality Ratio (IQR) or special case of dual total
            # correlation
            return mutual_info / (
                self.entropy(0, base=base)
                + self.entropy(1, base=base)
                - mutual_info
            )

        raise ValueError(f'Unexpected value for `normalized`: {normalized}')

    def entropy(self, axis, base=None):
        """Returns the entropy of either the predictions or actual values."""
        if axis == 0 or axis == 'pred' or axis == 'predicted':
            marginal = self.mat.sum(0) / self.mat.sum()
        elif axis == 1 or axis == 'actual':
            marginal = self.mat.sum(1) / self.mat.sum()
        else:
            raise ValueError(f'Unexpected value for `axis`: {axis}')
        return entropy(marginal, base=base)

    # TODO Reduction of classes including two class subsets to obtain
    # Binarization
    #   e.g. given set of labels as known and set of labels as unknown return
    #   binary confusion matrix of knowns vs uknowns

    # TODO combination of different confusion matrices objects into one.

    # TODO Visualizations
    # ROC
    # TOC

    def heatmap(
        self,
        filepath=None,
        overwrite=False,
        **kwargs,
    ):
        """Confusion matrix visualized as a heat map using plotly."""
        if self.labels is None:
            labels = np.arange(len(self.mat))
        else:
            labels = self.labels

        # Provide more informative (overridable) default layout for the heatmap
        layout_kwargs = dict(
            title = '<b>Confusion Matrix</b>',
            xaxis = {'side': 'top'},
            yaxis = {'autorange': 'reversed'},
        )
        # If layout override exists, then update defaults
        if 'layout' in kwargs:
            if isinstance(kwargs['layout'], dict):
                layout_kwargs.update(kwargs['layout'])
                kwargs['layout'] = go.Layout(**layout_kwargs)
            else:
                layout_kwargs.update(kwargs['layout'])
                kwargs['layout'] = go.Layout(**layout_kwargs).update(
                    **kwargs['layout'],
                )
        else:
            kwargs['layout'] = go.Layout(**layout_kwargs)


        fig = go.Figure(
            data=go.Heatmap(
                z=self.mat,
                x=labels,
                y=labels,
                **kwargs.pop('data', {}),
            ),
            **kwargs,
        )

        if filepath is None:
            return fig

        # Otherwise save plot to filepath
        fig.write_image(create_filepath(filepath, overwrite=overwrite))

    def save(
        self,
        filepath,
        conf_mat_key='confusion_matrix',
        overwrite=False,
        *args,
        **kwargs,
    ):
        """Saves the current confusion matrix to the given filepath."""
        ext = os.path.splitext(filepath)[-1]

        if ext not in {'.csv', '.tsv', '.hdf5', '.h5'}:
            raise TypeError(' '.join([
                'Expected file extention: ".csv", ".tsv", ".hdf5", or',
                f'".h5"; not  `{ext}`',
            ]))

        filepath = create_filepath(filepath, overwrite=overwrite)

        if ext == '.csv':
            pd.DataFrame(self.mat, columns=self.labels).to_csv(
                filepath,
                index=False,
                *args,
                **kwargs,
            )
        elif ext == '.tsv':
            pd.DataFrame(self.mat, columns=self.labels).to_csv(
                filepath,
                index=False,
                sep='\t',
                *args,
                **kwargs,
            )
        else: # HDF5 elif ext in {'.hdf5', '.h5'}:
            with h5py.File(filepath, 'r') as h5f:
                h5f['labels'] = self.labels.astype(np.string_)
                h5f[conf_mat_key] = self.mat

    @staticmethod
    def load(
        self,
        filepath,
        sep=',',
        filetype=None,
        names=None,
        conf_mat_key='confusion_matrix',
        *args,
        **kwargs,
    ):
        """Convenience function that loads the confusion matrix from the given
        filepath in given common formats. Loading from CSV relies on
        pandas.read_csv(*args, **kwargs).
        """
        if filetype is None:
            # Infer filetype from filepath if filetype is not given
            parts = filepath.rpartition('.')

            if not parts[0]:
                raise ValueError(' '.join([
                    'filetype is `None` and no file extention present in',
                    'given filepath.',
                ]))

            if parts[2] and os.path.sep in parts[2]:
                raise ValueError(' '.join([
                    'filetype is `None` and no file extention present in',
                    'given filepath.',
                ]))

            filetype = parts[2].lower()
            if filetype == 'csv':
                sep = ','
                #filetype = 'csv'
            elif filetype == 'tsv':
                sep = '\t'
                #filetype = 'tsv'
            elif filetype == 'hdf5' or filetype == 'h5':
                filetype = 'hdf5'
            else:
                raise ValueError(' '.join([
                    'filetype is `None` and file extention present in',
                    'given filepath is not "csv", "tsv", "hdf5", or "h5".',
                ]))
        elif isinstance(filetype, str):
            # lowercase filetype and check if an expected extension
            filetype = filetype.lower()
            if filetype not in {'csv', 'tsv', 'hdf5', 'h5'}:
                raise TypeError(' '.join([
                    'Expected filetype to be a str: "csv", "tsv", "hdf5", or',
                    f'"h5", not `{filetype}`',
                ]))
        else:
            raise TypeError(' '.join([
                'Expected filetype to be a str: "csv", "tsv", "hdf5", or',
                f'"h5", not of type `{type(filetype)}`',
            ]))

        if filepath == 'csv' or 'tsv':
            loaded_conf_mat = pd.read_csv(
                filepath,
                sep=sep,
                names=names,
                *args,
                **kwargs,
            )

            assert (
                len(loaded_conf_mat.shape) == 2
                and loaded_conf_mat.shape[0] == loaded_conf_mat.shape[1]
            )

            if names is None:
                labels = np.array(loaded_conf_mat.columns)
            else:
                labels = None
        else:  # HDF5
            with h5py.File(filepath, 'r') as h5f:
                if 'labels' in h5f.keys():
                    if names is not None:
                        logger.warning(
                            '`names` is provided while "labels" exists in the '
                            'hdf5 file! `names` is prioritized of the labels '
                            'in hdf5 file.'
                        )
                        labels = names
                    else:
                        labels = h5f['labels'][:]

                loaded_conf_mat = h5f[conf_mat_key][:]

        return ConfusionMatrix(loaded_conf_mat, labels=labels)


class ConfusionMatrices(ConfusionMatrix):
    """Multiple confusion matrices with an arbitrary tensor serving as the
    indices, where each index is a confusion matrix. The indexing tensor may
    consist of nominal values, thus using the NominalDataEncoder.

    This extends ConfusionMatrix and is a slice of the sparse ConfusionTensor
    if each of the different confusion matrices were different discrete random
    variables being compared.
    """
    def __init__(self):
        raise NotImplementedError()


class ConfusionTensor(ConfusionMatrix):
    """A Confusion Tensor of M dimensions all of length N, resulting in the
    shape `[N_1, N_2, ..., N_M]`.
    """
    def __init__(self):
        raise NotImplementedError()

    # TODO All the above measures need generalized such that they are applied
    # to a matrix slice. The idea is to generalize them for this Tensor or a
    # slice of this Tensor, thus needing to work on sparse np.ndarrays and
    # easily apply to both cases, w/ this possibly being the Parent of
    # ConfusionMatrix and ConfusionMatrices.



"""Example showing how to use exputils.io's initial argparser."""
import logging

from exputils import io


def add_example_args(parser):
    example = parser.add_argument_group(
        'example',
        'Example args grouped together from adding custom args.',
    )

    example.add_argument(
        '--dir_path',
        help=' '.join([
            'The filepath to the directory containing either the separate',
            'results directories of a kfold experiment, or the root directory',
            'containing all kfold experiments to be loaded.',
        ]),
        default='./',
        dest='example.dir_path',
    )

    example.add_argument(
        '--summary_name',
        help=' '.join([
            'The relative filepath from the eval fold directory to the',
            'summary JSON file.',
        ]),
        default='summary.json',
        dest='example.summary_name',
    )


def add_custom_args(parser):
    add_example_args(parser)

    parser.add_argument(
        '--save_labels',
        action='store_true',
        help='Save the labels in the JSON of the predictions.'
    )

    parser.add_argument(
        '--pie_cake',
        action='store_true',
        help='Example of how NestedNamespaces recurse based on given dest',
        dest='so.much.pie.cake',
    )


if __name__ == '__main__':
    args = io.parse_args(custom_args=add_custom_args)

    logging.info('It ran.')



"""Tests of the compiler-like checking of docstrings to ensure they parse
styles as they are expected and includes examples of out-of-style
docstrings.

The fixtures are the expected parsed tokens from docstr.parse.
"""
from collections import OrderedDict
from copy import copy, deepcopy
import os
from string import ascii_uppercase, ascii_lowercase

from bidict import OrderedBidict
import pytest
import numpy as np

from exputils.data.labels import NominalDataEncoder as NDE


@pytest.fixture
def example_labels():
    return tuple(
        [str(i) for i in range(10)]
        + list(ascii_uppercase)
        + list(ascii_lowercase)
        + ['noxious']
    )


@pytest.fixture
def example_labels_file(request):
    return os.path.join(request.fspath.dirname, 'example_labels.txt')


@pytest.fixture
def example_labels_bidict(example_labels):
    return OrderedBidict({key: enc for enc, key in enumerate(example_labels)})


@pytest.mark.dependency(name='object_basics', scope='session')
#@pytest.mark.incremental
class TestObjectBasics:
    """The fundamental parsing of functions including defaults and choices."""
    def test_default_init_and_eq(self, example_labels, example_labels_bidict):
        nde_1 = NDE(example_labels)

        argsorted = np.unique(example_labels, return_index=True)[1]

        # Test the argsorted_keys given these example labels
        assert (nde_1._argsorted_keys == argsorted).all()

        # Test the default attributes
        assert nde_1.pos_label == 1
        assert nde_1.neg_label == 0
        assert nde_1.sparse_output is False
        assert nde_1._unknown_key is None

        # Check the order of example labels in the encoder, and check list conv
        assert tuple(nde_1.encoder) == example_labels
        assert tuple(nde_1) == example_labels
        assert nde_1.encoder == example_labels_bidict

        # Test the default properties
        assert nde_1.unknown_key is None
        assert nde_1.unknown_idx is None
        assert nde_1.shift is 0

        # Test equality check `==`, `!=`, `is`, `is not` between NDEs.
        nde_2 = NDE(example_labels)

        # Repeat tests on nde_2 as sanity check for any global changes in init
        # Test the argsorted_keys given these example labels
        assert (nde_2._argsorted_keys == argsorted).all()

        # Test the default attributes
        assert nde_2.pos_label == 1
        assert nde_2.neg_label == 0
        assert nde_2.sparse_output is False
        assert nde_2._unknown_key is None

        # Check the order of example labels in the encoder, and check list conv
        assert tuple(nde_2.encoder) == example_labels
        assert tuple(nde_2) == example_labels
        assert nde_2.encoder == example_labels_bidict

        # Test the default properties
        assert nde_2.unknown_key is None
        assert nde_2.unknown_idx is None
        assert nde_2.shift is 0

        # Test comparison of NDEs
        assert nde_1 == nde_2
        assert not (nde_1 != nde_2)
        assert not (nde_1 is nde_2)
        assert nde_1 is not nde_2

        # TODO test error on ignore_dups = False
        # TODO test no error on ignore_dups = True

    def test_load_save(self, example_labels, example_labels_file):
        assert NDE.load(example_labels_file) == NDE(example_labels)

    def test_copy(self, example_labels):
        nde_src = NDE(example_labels)
        nde_copy = copy(nde_src)
        assert nde_src == nde_copy

        # Change an object within one, witness change in the other
        nde_src._argsorted_keys[0] = -100
        assert nde_src == nde_copy

        nde_copy._argsorted_keys[-1] = -200
        assert nde_src == nde_copy

        # Change a literal within one, witness no change in the other
        tmp = nde_src.pos_label
        nde_src.pos_label = 100
        assert nde_src != nde_copy
        nde_src.pos_label = tmp

        nde_copy.pos_label = 100
        assert nde_src != nde_copy

    def test_deepcopy(self, example_labels):
        nde_src = NDE(example_labels)
        nde_copy = deepcopy(nde_src)
        assert nde_src == nde_copy

        # Change an object within one, witness no change in the other
        tmp = nde_src._argsorted_keys[0]
        nde_src._argsorted_keys[0] = -100
        assert nde_src != nde_copy
        nde_src._argsorted_keys[0] = tmp
        assert nde_src == nde_copy

        tmp = nde_copy._argsorted_keys[-1]
        nde_copy._argsorted_keys[-1] = -200
        assert nde_src != nde_copy

    def test_encoder_bidict_methods(self, example_labels):
        """Methods that are pass throughs to the bidict encoder."""
        nde = NDE(example_labels)

        # Double Unders
        # Test len
        assert len(nde) == len(nde.encoder)

        # Test contains, getitem, dict method: get
        for i, label in enumerate(example_labels):
            assert label in nde
            assert nde[label] == i
            assert nde.get(label) == i

        # Test iter and reversed
        assert tuple(iter(nde)) == example_labels
        assert tuple(reversed(nde)) == tuple(reversed(example_labels))

        # Dict methods: keys, values, items, get
        assert nde.keys() == nde.encoder.keys()
        assert nde.values() == nde.encoder.values()
        assert nde.items() == nde.encoder.items()
        assert nde.get('pie', None) is None

        # Bidict properties / methods
        assert nde.inverse is nde.encoder.inverse
        assert nde.inv is nde.encoder.inv
        assert nde.inv is nde.encoder.inverse


@pytest.mark.dependency(name='encoder_knowns', depends=['object_basics'])
class TestLabelEncoder:
    @pytest.mark.parametrize(
        'one_hot,shift,unknown_key',
        [
            (False, 0, None),
            (True, 0, None),
            (False, 20, None),
            (True, 20, None),
            # label '0' treated as unknown
            (False, 0, '0'),
            (True, 0, '0'),
            (False, 20, '0'),
            (True, 20, '0'),
            # Unknown added at first idx
            (False, 0, 'unknown'),
            (True, 0, 'unknown'),
            (False, 20, 'unknown'),
            (True, 20, 'unknown'),
        ],
    )
    def test_encode_decode(self, one_hot, shift, unknown_key, example_labels):
        nde = NDE(example_labels, shift=shift, unknown_key=unknown_key)

        if unknown_key not in {None, '0'}:
            example_labels = tuple(['unknown'] + list(example_labels))

        ref_labels = np.array(example_labels)
        n_labels = len(example_labels)

        if one_hot:
            ref_enc = np.eye(n_labels)
            decode_axes = [-1] * 10
            decode_axes_x16 = [-1] * 4
        else:
            ref_enc = np.arange(n_labels) + shift
            decode_axes = [None] * 10
            decode_axes_x16 = [None] * 4

        # Check encoding of a list and tuple
        assert nde.shift == shift
        assert (nde.encode(example_labels, one_hot) == ref_enc).all()
        assert (nde.encode(list(example_labels), one_hot) == ref_enc).all()

        # Check np.ndarray encoding and decoding of different shapes
        if unknown_key not in {None, '0'}:
            shapes = (
                [n_labels],
                [1, n_labels],
                [1, 1, 1, n_labels, 1],
                [8, 8],
                [32, 2],
                [2, 16, 2],
                [4, 16],
                [16, 4],
                [2, 2, 4, 4],
                [2, 2, 2, 2, 4],
            )
        else:
            shapes = (
                [n_labels],
                [1, n_labels],
                [1, 1, 1, n_labels, 1],
                [3, 21],
                [21, 3],
                [7, 3, 3],
                [3, 7, 3],
                [3, 3, 7],
                [9, 7],
                [7, 9],
            )
        for i, shape in enumerate(shapes):
            encoded = nde.encode(ref_labels.reshape(shape), one_hot)
            enc_shape = shape + [n_labels] if one_hot else shape
            assert (encoded == ref_enc.reshape(enc_shape)).all()
            assert (
                nde.decode(encoded, decode_axes[i])
                == ref_labels.reshape(shape)
            ).all()

        if unknown_key != '0':
            return

        # Check np.ndarray encoding and decoding of larger dimensions
        ref_labels_x16 = np.concatenate([ref_labels] * 16)
        ref_enc_x16 = np.concatenate([ref_enc] * 16)
        shapes_x16 = (
            [2, 8, 3, 7, 3],
            [4, 4, 3, 7, 3],
            [2, 2, 2, 2, 3, 7, 3],
            [1, 16, 3, 7, 3],
        )
        for i, shape in enumerate(shapes_x16):
            encoded = nde.encode(ref_labels_x16.reshape(shape), one_hot)
            enc_shape = shape + [n_labels] if one_hot else shape
            assert (encoded == ref_enc_x16.reshape(enc_shape)).all()
            assert (
                nde.decode(encoded, decode_axes_x16[i])
                == ref_labels_x16.reshape(shape)
            ).all()

    def test_shift_encoding(self, example_labels):
        """Tests the shift_encoding method for positive and negative shifts."""
        nde = NDE(example_labels)
        assert nde.shift == 0

        nde_shifted = NDE(example_labels, shift=20)
        assert nde_shifted.shift == 20

        nde.shift_encoding(20)

        assert nde == nde_shifted

        nde = NDE(example_labels)
        nde_shifted.shift_encoding(-20)
        assert nde == nde_shifted

    def test_append_pop(self, example_labels):
        """Tests the append and pop methods."""
        nde = NDE(example_labels)
        nde_copy = deepcopy(nde)
        nde_appended = NDE(list(example_labels) + ['pie'])

        assert nde != nde_appended
        assert nde_copy != nde_appended
        assert nde == nde_copy

        # Test an individual key first.
        nde.append('pie')

        assert nde != nde_copy
        assert nde == nde_appended

        nde.pop('pie')

        assert nde == nde_copy
        assert nde != nde_appended

        # Test multiple keyes
        nde_appended = NDE(list(example_labels) + ['pie', 'cake', 'cheese'])
        nde.append(['pie', 'cake', 'cheese'])
        assert nde == nde_appended
        assert nde != nde_copy

        nde_appended_copy = deepcopy(nde)
        assert nde == nde_appended_copy

        # TODO Include pop multiple keys? nde.pop(['pie', 'cake', 'cheese'])
        for key in ['pie', 'cake', 'cheese']:
            nde.pop(key)
        assert nde == nde_copy

        for key in ['pie', 'cake', 'cheese']:
            nde.append(key)
        assert nde == nde_appended

        # Test removal of a key value pair using the encoding.
        for i in range(1, 4):
            nde.pop(66 - i, True)
        assert nde == nde_copy

        # Test removal of a key in the middle
        ex_labels_miss_A = list(example_labels)
        del ex_labels_miss_A[10]
        nde_missing_A = NDE(ex_labels_miss_A)

        nde.pop('A')
        nde == nde_missing_A

        # Test shift update after pop
        nde = NDE(example_labels, shift=12)
        assert nde.shift == 12
        nde.pop('0')
        assert nde.shift == 12

    #@pytest.mark.xfail
    #TODO def test_reorder(self, example_labels):
    #    assert False


#@pytest.mark.dependency(depends=['encoder_knowns'])
class TestUnknownLabel:
    def test_init_unknown_key_given_in_ordered_keys(self, example_labels):
        # At beginning
        nde = NDE(['unknown'] + list(example_labels), unknown_key='unknown')
        assert nde.unknown_key == 'unknown'
        assert nde.unknown_idx == 0

        nde = NDE(
            ['unknown'] + list(example_labels),
            unknown_key='unknown',
            unknown_idx=0,
        )
        assert nde.unknown_key == 'unknown'
        assert nde.unknown_idx == 0

        # At end
        nde = NDE(list(example_labels) + ['unknown'], unknown_key='unknown')
        assert nde.unknown_key == 'unknown'
        assert nde.unknown_idx == len(example_labels)

        nde = NDE(
            list(example_labels) + ['unknown'],
            unknown_key='unknown',
            unknown_idx=len(example_labels),
        )
        assert nde.unknown_key == 'unknown'
        assert nde.unknown_idx == len(example_labels)

        # In middle
        nde = NDE(example_labels, unknown_key='A')
        assert nde.unknown_key == 'A'
        assert nde.unknown_idx == 10

        # Expects ValueError
        raised_error = False
        try:
            nde = NDE(example_labels, unknown_key='A', unknown_idx=22)
        except ValueError:
            raised_error = True
        assert raised_error

    def test_init_unknown_key_given_not_in_ordered_keys(self, example_labels):
        # At beginning
        nde = NDE(example_labels, unknown_key='unknown')
        assert nde.unknown_key == 'unknown'
        assert nde.unknown_idx == 0

        # At end
        nde = NDE(example_labels, unknown_key='unknown', unknown_idx=-1)
        assert nde.unknown_key == 'unknown'
        assert nde.unknown_idx == len(example_labels)

        # In middle
        assert nde['A'] == 10
        nde = NDE(example_labels, unknown_key='unknown', unknown_idx=10)
        assert nde.unknown_key == 'unknown'
        assert nde.unknown_idx == 10
        assert nde['A'] == 11

        # Expects ValueError
        raised_error = False
        try:
            nde = NDE(example_labels, unknown_key='unknown', unknown_idx=220)
        except TypeError:
            raised_error = True
        assert raised_error

    def test_shift(self, example_labels):
        nde = NDE(example_labels, unknown_key='unknown')
        assert nde.unknown_key == 'unknown'
        assert nde.unknown_idx == 0
        assert nde.shift == 0

        nde_copy = deepcopy(nde)
        assert nde == nde_copy

        nde_shifted = NDE(example_labels, shift=20, unknown_key='unknown')
        assert nde_shifted.shift == 20
        assert nde_shifted.unknown_key == 'unknown'
        assert nde_shifted.unknown_idx == 20

        nde.shift_encoding(20)

        assert nde == nde_shifted

        nde_shifted.shift_encoding(-20)
        assert nde_copy == nde_shifted

    def test_pop(self, example_labels):
        nde = NDE(example_labels, unknown_key='unknown')
        nde_knowns = NDE(example_labels)
        assert nde.unknown_key == 'unknown'
        assert nde.unknown_idx == 0

        nde.pop(nde.unknown_key)
        assert nde.unknown_key is None
        assert nde.unknown_idx is None
        assert nde == nde_knowns

    def test_pop_shift(self, example_labels):
        nde = NDE(example_labels, unknown_key='unknown', shift=12)
        nde_knowns = NDE(example_labels, shift=12)
        assert nde.unknown_key == 'unknown'
        assert nde.unknown_idx == 12

        assert nde.pop(nde.unknown_key) == 12
        assert nde.unknown_key is None
        assert nde.unknown_idx is None
        assert nde == nde_knowns
        assert nde.shift == 12

    #@pytest.mark.xfail
    #def test_reorder(self):
    #    assert False

    # TODO test unknown encoding of unknown labels. and decoding.



"""Abstract or Generic classes for predictors of different types. Intended to organize
creation of predictors.
"""
from abc import ABC, abstractmethod
import pickle
import gzip

import numpy as np

from exputils.data.labels import NominalDataEncoder
from exputils.io import create_filepath


class Stateful(object):
    """Generic class for a stateful object who needs save and load methods."""
    def save(self, filepath, overwrite=False):
        """Given filepath, save the current state of the object."""
        # TODO attempt generic compression and saving of python object using
        # more efficent compression for numerical data w/o the need for
        # serialization, e.g. hdf5 or numpy npz. Otherwise, fallback to
        # compressed pickle (unnecessary in general case, consider separate
        # method to call for when it is necessary, e.g. complicated models)

        # TODO if diff saving used, chose via file ext, and set ext if none
        # after attempting above and determining which is used.

        # Generic save using pickle with gzip compression
        with gzip.open(create_filepath(filepath, overwrite), 'wb') as openf:
            openf.write(pickle.dumps(self))

        # TODO add option to save params to json/yaml . . . ? (opt. compress)
        #   At least add to dict (vars) and __str__ version

        # TODO make the different generic saves/loads be their own functions
        # that are called here and then may be called elsewhere too.

    @staticmethod
    def load(filepath):
        """Given filepath, load the saved state of the object."""
        # Generic load using pickle
        with gzip.open(filepath, 'rb') as openf:
            return pickle.load(self, openf.read())


class Predictor(Stateful):
    """Abstract class for predictors."""
    @abstractmethod
    def predict(self, features):
        """Given the current state of the predictor, predict the labels"""

        # TODO predict in batches
        raise NotImplementedError()


class SupervisedLearner(Predictor):
    """Abstract class for supervised learning predictors."""

    # TODO __init__ either sets random state or setups up random state to be
    # created when first fit occurs (implies unable to predict in this case).
    # assumes stochastic learning.

    @abstractmethod
    def fit(self, features, labels):
        """Given the current state of the predictor, continue training given
        the provided data. This uses the existing state of the predictor.
        """

        # TODO fit in batches
        # TODO fit incrementally
        # TODO parameterize init, fit, and predict st they may be provided args
        # & kwargs as configs
        raise NotImplementedError()


class SupervisedClassifier(SupervisedLearner):
    """Abstract class for supervised learning classifiers.

    Attributes
    ----------
    label_enc : NominalDataEncoder
    """

    # TODO __init__ either sets random state or setups up random state to be
    # created when first fit occurs (implies unable to predict in this case).

    def __init__(self, *args, **kwargs):
        """Initialize a supervised classifier with a nominal data encoder.

        Args
        ----
        see NominalDataEncoder.__init__
        """
        self.label_enc = NominalDataEncoder(*args, **kwargs)

    @property
    def labels(self):
        return np.array(self.label_enc.encoder)

# TODO numpy, python, tensorflow, torch, etc. RandomState class that serves as
# a single randomstate object and updates whenever it is used by any of these
# libs, keeping it consistent and easy to manage in a cross lib project.
# Someone may have done this already.



"""
A script for generating a set of random seeds to use for more reproducible
results.
"""
import argparse
import random
import sys


def generate(number_of_seeds, seed=None):
    """Uses pythons random integer generator to create a set of unique random
    seeds.

    Parameters
    ----------
    number_of_seeds : int
        number of seeds to generate.
    seed : int, optional
        The random seed to intialize the generator with.

    Returns
    -------
        A list of integers.
    """
    if seed is not None:
        if isinstance(seed, int):
            random.seed(seed)
        else:
            raise TypeError('The inital seed is expected to be an integer.')

    random_seeds = set()
    # max_int = sys.maxsize
    max_int = 2**32 - 1
    i = 0
    while i < number_of_seeds:
        random_seed = random.randint(0, max_int)
        if random_seed not in random_seeds:
            random_seeds.add(random_seed)
            i += 1

    return random_seeds


def save_to_file(seeds, filepath):
    """Saves the random seeds as a file at the fiven filepath

    Parameters
    ----------
    seeds : numpy.array
        A 1D numpy.array of random seeds
    """
    with open(filepath, 'w') as f:
        for seed in seeds:
            f.write('%d\n' % seed)


def parse_args():
    parser = argparse.ArgumentParser(
        description='Generate unique random seeds and save to a file.',
    )
    parser.add_argument(
        'number_of_seeds',
        type=int,
        help='The number of seeds to be generated by calling this script.',
    )
    parser.add_argument(
        '-s',
        '--seed',
        default=None,
        type=int,
        help='An integer to serve as the initial seed to the generator.',
    )
    parser.add_argument(
        '-o',
        '--output_filepath',
        default=None,
        help='The filepath to save the random seeds file to.',
    )

    args = parser.parse_args()

    if args.output_filepath is None:
        args.output_filepath = f'random_seeds_count-{args.number_of_seeds}.txt'

    return args


if __name__ == '__main__':
    args = parse_args()
    seeds = generate(args.number_of_seeds, args.seed)
    save_to_file(seeds, args.output_filepath)



"""Convenience functions for profiling python code. Please remember that both
cProfile and profile packages exist in the Python standard library. They may be
used to profile the compute times of all function calls within a program. These
are for convenience and simplicity, especially when the recorded times will be
used by the program, but the results from cProfile may also be used similarly.
"""
from datetime import datetime
from contextlib import contextmanager
import logging
from time import time, process_time

class TimeData(object):
    """Contains the different kinds of time data when timing content.

    Attributes
    ----------
    start_time : float
        Start time in seconds.
    end_time : float
        End time in seconds.
    elapsed_time : float
        Elapsed time between start and end in seconds.
    start_datetime : float
    end_datetime : float
    elapsed_datetime : float
    """
    def __init__(self, time_method=time):
        self.time = time_method
        # TODO ensure this is as low overhead as possible for timing

    @property
    def start_time(self):
        return self._start_time

    @property
    def end_time(self):
        return self._end_time

    @property
    def elapsed_time(self):
        return self._end_time - self._start_time

    @property
    def start_datetime(self):
        return self._start_datetime

    @property
    def end_datetime(self):
        return self._end_datetime

    @property
    def elapsed_datetime(self):
        return self._end_datetime - self._start_datetime

    def start(self):
        """Record the starting time. The timers are ordered by grainularity."""
        self._start_datetime = datetime.now()
        self._start_time = time()

    def end(self):
        """Record the ending time. The timers are ordered by grainularity."""
        self._end_time = time()
        self._end_datetime = datetime.now()


def time_func(func, *args, **kwargs):
    """Given a function or callable, times its wall runtime in seconds."""
    timedata = TimeData()

    timedata.start()
    results = func(*args, **kwargs)
    timedata.end()

    return timedata


# TODO consider adding wall_time and making these two partial funcs of
# time_func(time_method, func, *args, **kwargs)
#def wall_time_func()


def proc_time_func(func, *args, **kwargs):
    """Given a function or callable, times its process runtime in seconds."""
    timedata = TimeData(process_time)

    timedata.start()
    results = func(*args, **kwargs)
    timedata.end()

    return timedata


@contextmanager
def log_time(start_message, end_message=None, time_method=time):
    """Logs the wall time of the content within the `with` block and logs it.

    Parameters
    ----------
    start_message : str
        Message to be written to the log before timing begins.
    end_message : str, optional
        Message to be written to the log after timing ends.
    time_method : callable
        A callable that operate like python's `time.time()`
    """
    # TODO implement writing to a specific, different log file than current one
    logging.info(start_message)

    start = time_method()
    yield
    stop = time_method()

    if end_message is not None:
        logging.info(end_message)

    logging.info("Elapsed time: %d", stop - start)

# TODO consider, if feasible, profiling cpu, gpu, memory usage.



"""Ordered confusion matricies for calculating top-k measures."""
import os

import h5py
import numpy as np
ma = np.ma

from exputils.data.labels import NominalDataEncoder as NDE
from exputils.data import ConfusionMatrix
from exputils.io import create_filepath

import logging
logger = logging.getLogger(__name__)


def scatter_nd_numpy(indices, updates, shape, target=None):
    """Equivalent to tf.scatter_nd()."""
    # TODO match tf scatter_nd interface
    # This does not work for recurring indices, which is necessary for CMs
    if target is None:
        target = np.zeros(shape, dtype=updates.dtype)
    np.add.at(
        target,
        tuple(indices.reshape(-1, indices.shape[-1]).T),
        updates.ravel(),
    )
    if target is None:
        return target


class OrderedConfusionMatrices(object):
    """A Tensor of KxCxC where K dimension represents the ordered confusion,
    and max(K) == C with each individual CxC matrix along the K dimenison is
    the ordered slice of the ordered predictions. Looking at one slice of the
    internal confusion tensor is NOT the top-k confusion matrix, but rather the
    confusion matrix when looking at n-th highest probable class. This may be
    used to obtain the top-k measures.

    Attributes
    ----------
    tensor : np.array
    label_enc : NominalDataEncoder

    Notes
    -----
    The ordered k confusion matrices loses the sample informaiton where the
    sample is the probability vector for the predicted class, which contains
    the order and relation of predicted classes for that sample. This only
    contains the confusion matrix for the k-th confusion matrix where the first
    k=1 confusion matrix is the standad confusion matrix (top-1). The second
    confusion matrix looks at the second most probable class in the predicted
    probability vectors, and so on for the values of k.
    """
    def __init__(
        self,
        targets,
        preds=None,
        labels=None,
        top_k=None,
        #axis=1,
        sort_labels=False,
        targets_idx=True,
        #weights=None,
    ):
        """Calculates the top-k confusion matrix of occurrence from the pairs.

        Args
        ----
        targets : array-like
            Vector of labels with length N.
        preds : array-like
            shape [N, C] for C classes, assumed to match labels
        labels : list = None
        top_k : int = None
        targets_idx : bool = True
            If True (default), casts the targets to their index encoding.
            The calculation of the top k-th confusion matrix is dependent upon
            the use of integer index encoding representaiton of the symbols.
        """
        # TODO generalize about axis, esp in the k loop of unique indices
        #axis = 1

        # NominalDataEncoder of labels to track symbol to the index
        if preds is None:
            if (
                isinstance(targets, np.ndarray)
                and len(targets.shape) == 3
                and targets.shape[1] == targets.shape[2]
            ):
                self.tensor = targets
                if top_k is None:
                    self.top_k = targets.shape[0]
                else:
                    raise ValueError(
                        'top_k given when targets is the ordered confusoin '
                        'matrices.'
                    )

                if labels is None:
                    raise ValueError(
                        'OrderedConfusionMatrices without labels!'
                    )
                else:
                    self.label_enc = NDE(labels)
        else:
            if top_k is None:
                top_k = 1
            else:
                assert(isinstance(top_k, int))

            if labels is None:
                self.label_enc = NDE(list(set(targets) | set(preds)))
            else:
                self.label_enc = NDE(labels)

            n_classes = len(self.label_enc)
            assert(preds.shape[1] == n_classes)

            # Get top-k predictions, assuming prediction
            top_preds = np.argsort(preds, axis=1)[:, ::-1][:, :top_k]

            # Cast both targets and preds to their args for a confusion matrix
            # based on label_enc
            if not targets_idx:
                targets = self.label_enc.encode(targets)

            # TODO decide if supporting weights is even necessary as this is counts
            #    if weights is not None:
            #        unique_idx_n *= weights
            self.tensor = get_cm_tensor(
                targets,
                top_preds,
                top_k,
                n_classes,
                axis=0,
            )

    def __add__(self, other):
        """Necessary for checking the changes over increments."""
        raise NotImplementedError()

    def get_conf_mat(self):
        """Returns the top-1 ConfusionMatrix, the first matrix in tensor."""
        return ConfusionMatrix(self.tensor[0], labels=np.array(self.label_enc))

    # TODO obtain BinaryTopKConfusionMatrix / Tensor and its associated
    # measures.
    def get_per_class_binary_top(self, k, axis1=1, axis2=2):
        """Returns a matrix of [C, 3] for C classes and the outcome per class
        using top-k logic of correct prediction (true positive per class)
        versus the sum of predicting incorrectly (false positive per class).
        The third vector is the false negatives.
        """
        true_pos = self.tensor.diagonal(axis1=1, axis2=2).sum(0)
        non_diag = ma.masked_array(
            self.tensor,
            mask=np.stack([np.eye(k, dtype=bool)] * k),
        )
        false_pos = non_diag.sum(1)
        false_neg = non_diag.sum(2)
        #true_positives
        return np.hstack((true_pos, false_pos, false_neg))

    def accuracy(self, k=None):
        """Top-k accuracy. K is maxed by default if not given. k inclusive"""
        # TODO unit test this.
        assert(k is None or isinstance(k, int))
        if k is None or k == 1:
            return (
                self.tensor[0].diagonal().sum()
                / self.tensor[0].sum()
            )
        elif k < 1:
            raise ValueError('k is to be >= 1 or None, but given k < 1!')
        return (
            self.tensor[:k].diagonal(axis1=1, axis2=2).sum()
            / self.tensor[0].sum()
        )

    # TODO consider how top-k may relate to the discrete mutual information.
    # research and look up if any existing measures for this.
    def save(
        self,
        filepath,
        conf_mat_key='ordered_confusion_matrices',
        overwrite=False,
        *args,
        **kwargs,
    ):
        """Saves the current confusion matrix to the given filepath."""
        if not isinstance(filepath, str):
            raise NotImplementedError(
                'Only str filepaths are supported for saving.',
            )

        ext = os.path.splitext(filepath)[-1]

        if ext not in {'.hdf5', '.h5'}:
            raise TypeError(
                f'Expected file extention: ".hdf5", or ".h5"; not `{ext}`',
            )

        filepath = create_filepath(filepath, overwrite=overwrite)

        with h5py.File(filepath, 'w') as h5f:
            h5f['labels'] = np.array(self.label_enc).astype(np.string_)
            h5f[conf_mat_key] = self.tensor

    @staticmethod
    def load(
        filepath,
        labels=None,
        conf_mat_key='ordered_confusion_matrices',
    ):
        """Convenience function that loads the confusion matrix from the given
        filepath in given common formats. Loading from CSV relies on
        pandas.read_csv(*args, **kwargs).
        """
        ext = os.path.splitext(filepath)[-1]

        if ext not in {'.hdf5', '.h5'}:
            raise TypeError(
                f'Expected file extention: ".hdf5", or ".h5"; not `{ext}`',
            )

        with h5py.File(filepath, 'r') as h5f:
            if 'labels' in h5f.keys():
                if labels is not None:
                    logger.warning(
                        '`names` is provided while "labels" exists in the '
                        'hdf5 file! `names` is prioritized of the labels '
                        'in hdf5 file.'
                    )
                else:
                    labels = [label.decode() for label in h5f['labels'][:]]

            loaded_conf_mat = h5f[conf_mat_key][:]

        return OrderedConfusionMatrices(loaded_conf_mat, labels=labels)


def get_cm_tensor(targets, top_preds, top_k, n_classes, axis=0):
    # TODO make sparse numpy matrix/tensor w/ zero as fill.
    ordered_cms = np.zeros([top_k, n_classes, n_classes])
    for k in range(top_k):
        unique_idx, unique_idx_counts = np.unique(
            np.stack([targets, top_preds[:, [k]]], axis=axis),
            return_counts=True,
            axis=1,
        )
        np.add.at(
            ordered_cms[k],
            tuple(unique_idx.T.reshape(-1, unique_idx.T.shape[-1]).T),
            unique_idx_counts.ravel(),
        )
    return ordered_cms



"""Generic Dataset Loader intended to serve as the base for specific loders."""
import os

import pandas as pd

# TODO PostgreSQL config parser w/ psycopg2

# TODO this may not be a real generic, as it does not easily turn into all the
# other PyTorch Dataset classes...
# So perhaps some basic tools commonly used when making Dataset classes and
# data loaders would be best?

class GenericDataset(object):
    """Generic Dataset specification intended to be the backbone to all dataset
    specifications and loaders, across Tensorflow and Pytorch. Simply put, the
    generic code for loading from common data sources, such as csv, tsv, hdf5,
    and databases.
    """

    def __init__(self, filepath, *args, **kwargs):
        if isinstance(filepath, str):
            basename, ext = os.path.splitext(filepath)
            # TODO add generic support for HDF5 and .tar.gz for WebDataset???
            #   Perhaps separate these into their own base dataset clases?
            if ext not in {'csv', 'tsv', 'ini'}:
                raise ValueError(' '.join([
                    "Expected one of the filetypes:{'csv', 'tsv', 'ini'}, but",
                    f"recieved: {ext} at `filepath`: {filepath}",
                ]))
            # TODO load dataset from CSV or TSV

            # TODO perhaps, Save the column headers as a NamedTuple or
            # DataClass for an item description?
        #elif isinstance(filepath, PostgreSQLConfig)
        else:
            raise TypeError(
                '`filepath` expected to be type `str`, not {type(filepath)}.'
            )

    def __len__(self):
        raise NotImplementedError()

    def __getitem__(self):
        raise NotImplementedError()

        # TODO to implement slicing, the databackend must support it. This is
        # why there is IterableDataset in PyTorch. If a backend is akin to a
        # dict/map, then perhaps treat it as an OrderedDict, or make a more
        # specific generic subclass that will.
        # NOTE to have both iterable and map by id, you essentially want the
        # pandas.DataFrame, just efficient for I/O things

    def __add__(self, other):
        """Generic concatenation of datasets, similar to python lists."""
        raise NotImplementedError()

        # NOTE this is a bad idea, beacuase in the end, it is probably better
        # to treat this as a tensor, thus wanting to preserve math ops for
        # tensor math. So simply add general append, stack, etc... funcs or
        # methods.

    # TODO implement numpy style slicing??

    # TODO implement partitioning and kfold partitions as a Dataset method?

#class GenericIterDataset(GenericDataset):
#    """A generic iterable dataset."""

# TODO consider adding a Streaming dataset, or iterable I/O dataset that only
# reads in portions of the dataset from the source at a time, such as from a DB
# or from part of a file.
# If not using a Torch or TF dataset/loader, the consider Dask as a streaming
# alternative to Pandas for large data files (and parallel). Albeit, not an
# alternative to a database like PostgreSQL. However, since research code is a
# mess, its nice for the option to parallelie many csvs or tabular data.



"""The general Input / Output of experiments."""
import argparse
from copy import deepcopy
from datetime import datetime
import json
import logging
import os
import sys

import h5py
import numpy as np
import keras
import tensorflow as tf


def add_output_args(parser):
    """Adds the arguments for specifying how to save output."""
    parser.add_argument(
        '-o',
        '--output_dir',
        default='./',
        help='Filepath to the output directory.',
    )

    parser.add_argument(
        '-s',
        '--summary_path',
        default='summary/',
        help='Filepath appeneded to `output_dir` for saving the summaries.',
    )


def add_data_args(parser):
    """Adds the data arguments defining what data is loaded and used."""
    # NOTE may depend on model.parts
    data = parser.add_argument_group('data', 'Arguments pertaining to the '
        + 'data loading and handling.')

    # TODO add dataset id here, and have code expect that.

    data.add_argument(
        '-d',
        '--dataset_id',
        default='LabelMe',
        help='The dataset to use',
        choices=['LabelMe', 'FacialBeauty', 'All_Ratings'],
        #dest='' # TODO make the rest of the code expect this to be together ..?
        # mostly requires changing `predictors.py` as dataset_handler.load_dataset()
    )
    data.add_argument(
        'data.dataset_filepath',
        help='The filepath to the data directory',
        #dest='data.dataset_filepath',
    )

    # TODO add to data args and expect it in Data Classes.
    data.add_argument(
        '-l',
        '--label_src',
        default='majority_vote',
        help='The source of labels to use for training.',
        choices=['majority_vote', 'frequency', 'ground_truth', 'annotations'],
        #dest='data.label_src',
    )


def add_mle_args(parser):
    mle = parser.add_argument_group('mle', 'Arguments pertaining to the '
        + 'Maximum Likelihood Estimation.')

    mle.add_argument(
        '--max_iter',
        default=10000,
        type=int,
        help='The maximum number of iterations for finding MLE.',
        dest='mle.max_iter',
    )

    mle.add_argument(
        '--num_top_likelihoods',
        default=1,
        type=int,
        help='The number of top MLEs to be saved for each distribution.',
        dest='mle.num_top_likelihoods',
    )

    mle.add_argument(
        '--const_params',
        default=None,
        nargs='+',
        type=str,
        help='The model\'s parameters to be kept constant throughout the '
            + 'estimate of the MLE.',
        dest='mle.const_params'
    )

    mle.add_argument(
        '--alt_distrib',
        action='store_true',
        help=' '.join([
            'Whether to use the alternate parameterization of the given',
            'distribution, if it exists (ie. mean and precision for the',
            'Dirichlet)',
        ]),
        dest='mle.alt_distrib',
    )

    # Tolerances
    mle.add_argument(
        '--tol_param',
        default=1e-8,
        type=float,
        help='The threshold of parameter difference to the prior parameters '
            + 'set before declaring convergence and terminating the MLE '
            + 'search.',
        dest='mle.tol_param',
    )
    mle.add_argument(
        '--tol_loss',
        default=1e-8,
        type=float,
        help='The threshold of difference to the prior negative log likelihood '
            + 'set before declaring convergence and terminating the MLE '
            + 'search.',
        dest='mle.tol_loss',
    )
    mle.add_argument(
        '--tol_grad',
        default=1e-8,
        type=float,
        help='The threshold of difference to the prior gradient set before '
            + 'declaring convergence and terminating the MLE search.',
        dest='mle.tol_grad',
    )

    mle.add_argument(
        '--tol_chain',
        default=3,
        type=int,
        help=' '.join([
            'The number of iterations that a tolerance must be surpassed in',
            'order to be considered as convergence. Default is 1, meaning as',
            'soon as the tolerance threshold is surpassed, it is considered',
            'to have converged. This is a soft chain of tolerances, meaning',
            'that the tally of number of surpassed tolerances only increments',
            'and decrements by one every iteration, staying within the range ',
            'of [0. tol_chain]. The tally does not reset to 0 after a single',
            'iteration of not surpassing the tolerance threshold.',
        ]),
        dest='mle.tol_chain',
    )

    # optimizer_args
    mle.add_argument(
        '--learning_rate',
        #default=1e-3,
        default=.8,
        type=float,
        help='A Tensor or a floating point vlaue. The learning rate.',
        dest='mle.optimizer_args.learning_rate',
    )
    mle.add_argument(
        '--beta1',
        default=0.9,
        type=float,
        help='A float value or a constant float tensor. The exponential decay '
            + 'rate for the 1st moment estimates.',
        dest='mle.optimizer_args.beta1',
    )
    mle.add_argument(
        '--beta2',
        default=0.999,
        type=float,
        help='A float value or a constant float tensor. The exponential decay '
            + 'rate for the 2nd moment estimates',
        dest='mle.optimizer_args.beta2',
    )
    mle.add_argument(
        '--epsilon',
        default=1e-08,
        type=float,
        help='A small constant for numerical stability. This epsilon is '
            + '"epsilon hat" in the Kingma and Ba paper (in the formula just '
            + 'before Section 2.1), not the epsilon in Algorithm 1 of the '
            + 'paper.',
        dest='mle.optimizer_args.epsilon',
    )
    mle.add_argument(
        '--use_locking',
        action='store_true',
        help='Use locks for update operations.',
        dest='mle.optimizer_args.use_locking',
    )

    # tb_summary_dir ?? handled by output dir? or summary dir

def add_model_args(parser):
    """Adds the model arguments for `predictors.py`."""
    model = parser.add_argument_group(
        'model',
        'Arguments of the model to be loaded, trained, or evaluated.',
    )

    model.add_argument(
        '-m',
        '--model_id',
        default='vgg16',
        help='The model to use',
        choices=['vgg16', 'resnext50'],
        dest='model.model_id',
    )
    model.add_argument(
        '-p',
        '--parts',
        default='labelme',
        help='The part of the model to use, if parts are allowed (vgg16)',
        choices=['full', 'vgg16', 'labelme'],
        dest='model.parts',
    )

    # Init / Load
    model.add_argument(
        '--crowd_layer',
        action='store_true',
        help='Use crowd layer in ANNs.',
        dest='model.init.crowd_layer',
    )
    model.add_argument(
        '--kl_div',
        action='store_true',
        help='Uses Kullback Leibler Divergence as loss instead of Categorical '
            + 'Cross Entropy',
        dest='model.init.kl_div',
    )

    # TODO consider adding into model or putting into general (non-grouped) args
    # allow to be given a str
    parser.add_argument(
        '-r',
        '--random_seeds',
        default=None,
        nargs='+',
        #type=int,
        #type=multi_typed_arg(int, str),
        help='The random seed to use for initialization of the model.',
    )

    # Train
    model.add_argument(
        '-b',
        '--batch_size',
        default=32,
        type=int,
        help='The size of the batches in training.',
        dest='model.train.batch_size',
    )
    model.add_argument(
        '-e',
        '--epochs',
        default=1,
        type=int,
        help='The number of epochs.',
        dest='model.train.epochs',
    )

    model.add_argument(
        '-w',
        '--weights_file',
        default=None,
        help='The file containing the model weights to set at initialization.',
        dest='model.init.weights_file',
    )


def add_sjd_args(parser):
    """Adds the test SJD arguments to the argparser."""
    sjd = parser.add_argument_group(
        'sjd',
        'Arguments pertaining to tests evaluating the'
        + 'SupervisedJointDistribution in fitting simulated data.',
    )

    sjd.add_argument(
        '--target_distrib',
        type=multi_typed_arg(
            str,
            json.loads,
        ),
        help=' '.join([
            'Either a str identifer of a distribution or a dict with',
            '"distirb_id" as a key and the parameters of that distribution',
            'that serves as the target distribution.',
        ]),
        dest='sjd.target_distrib',
    )

    sjd.add_argument(
        '--transform_distrib',
        type=multi_typed_arg(
            str,
            json.loads,
        ),
        help=' '.join([
            'Either a str identifer of a distribution or a dict with',
            '"distirb_id" as a key and the parameters of that distribution',
            'that serves as the transform distribution.',
        ]),
        dest='sjd.transform_distrib',
    )

    sjd.add_argument(
        '--data_type',
        help='Str identifier of the type of data.',
        dest='sjd.data_type',
        default='nominal',
        choices=['nominal', 'ordinal', 'continuous'],
    )

    sjd.add_argument(
        '--independent',
        action='store_true',
        help=' '.join([
            'Indicates if the Supervised Joint Distribution\'s second random',
            'variable is independent of the first. Defaults to False.',
        ]),
        dest='sjd.independent',
    )

    # KNN desnity estimate parameters
    sjd.add_argument(
        '--knn_num_neighbors',
        type=float,
        help=' '.join([
            'A positive int for the number of neighbors to use in the K',
            'Nearest Neighbors density estimate of the transform pdf.',
        ]),
        default=10,
        dest='sjd.knn_num_neighbors',
    )

    sjd.add_argument(
        '--knn_num_samples',
        type=int,
        help=' '.join([
            'Number of samples to draw to approximate the transform pdf for ',
            'the K Nearest Neighbors density estimation.',
        ]),
        default=int(1e6),
        dest='sjd.knn_num_samples',
    )


def expand_mle_optimizer_args(args):
    """Put all mle-related args in a single dictionary."""
    if args.mle.optimizer_args and isinstance(args.mle.optimizer_args, NestedNamespace):
        args.mle.optimizer_args = vars(args.mle.optimizer_args)
    elif args.mle.optimizer_args:
        raise TypeError(f'`args.mle.optimizer` has expected type NestedNamespace, but instead was of type {type(args.mle.optimizer)} with value: {args.mle.optimizer}')


def expand_model_args(args):
    """Turns the init and train attributes into dicts."""
    if args.model.train and isinstance(args.model.train, NestedNamespace):
        args.model.train = vars(args.model.train)
    elif args.mle.train:
        raise TypeError(f'`args.mle.optimizer` has expected type NestedNamespace, but instead was of type {type(args.mle.optimizer)} with value: {args.mle.optimizer}')

    if args.model.init and isinstance(args.model.init, NestedNamespace):
        args.model.init = vars(args.model.init)
    elif args.mle.init:
        raise TypeError(f'`args.mle.optimizer` has expected type NestedNamespace, but instead was of type {type(args.mle.optimizer)} with value: {args.mle.optimizer}')


def parse_args(arg_set=None, custom_args=None, description=None):
    """Creates the args to be parsed and the handling for each.

    Parameters
    ----------
    arg_set : iterable, optional
        contains the additional argument types to be parsed.
    custom_args : function | callable, optional
        Given a function that expects a single argument to be
        `argparse.ArgumentParser`, this function adds arguments to the parser.

    Returns
    -------
    (argparse.namespace, None|int|list(ints))
        Parsed argumentss and random seeds if any.
    """
    if description:
        parser = argparse.ArgumentParser(description=description)
    else:
        parser = argparse.ArgumentParser(description='Run proof of concept.')

    add_logging_args(parser)
    add_hardware_args(parser)

    add_data_args(parser)
    add_output_args(parser)

    add_model_args(parser)
    add_kfold_cv_args(parser)

    if arg_set and 'mle' in arg_set:
        add_mle_args(parser)

    if arg_set and 'sjd' in arg_set:
        add_sjd_args(parser)

    # Add custom args
    if custom_args and callable(custom_args):
        custom_args(parser)

    args = parser.parse_args(namespace=NestedNamespace())

    # Set logging configuration
    numeric_level = getattr(logging, args.log_level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s' % args.log_level)
    if args.log_file is not None:
        dir_part = args.log_file.rpartition(os.path.sep)[0]
        os.makedirs(dir_part, exist_ok=True)
        logging.basicConfig(filename=args.log_file, level=numeric_level)
    else:
        logging.basicConfig(level=numeric_level)

    # Set the Hardware
    keras.backend.set_session(tf.Session(config=get_tf_config(
        args.cpu_cores,
        args.cpu,
        args.gpu,
    )))

    # TODO LOAD the randomseeds from file if it is of type str!!!
    if isinstance(args.random_seeds, list) and len(args.random_seeds) <= 1:
        if os.path.isfile(args.random_seeds[0]):
            raise NotImplementedError('Load the random seeds from file.')
        else:
            args.random_seeds[0] = int(args.random_seeds[0])
    elif isinstance(args.random_seeds, list):
        args.random_seeds = [int(r) for r in args.random_seeds]

    # TODO fix this mess here and its usage in `predictors.py`
    #if args.random_seeds and len(args.random_seeds) == 1:
    #    args.kfold_cv.random_seed = args.random_seeds[0]
    #    random_seeds = None
    #else:
    #    random_seeds = args.random_seeds

    #if args is not an int, draw from file.
    #if type(args.random_seeds, str) and os.path.isfile(args.random_seeds):

    #if type(args.random_seeds, list):
    #    print('random_seeds is a list!')

    expand_model_args(args)

    # expand early stopping args:
    args.kfold_cv.early_stopping = vars(args.kfold_cv.early_stopping)

    if arg_set and 'mle' in arg_set:
        expand_mle_optimizer_args(args)

        if arg_set and 'sjd' in arg_set:
            args.sjd.mle_args = vars(args.mle)
            args.sjd.n_jobs = args.cpu_cores

    #return args, random_seeds
    return args



"""Unofficial test. Just a quick ipython script test."""
from exputils.data.confusion_tensor import OrderedConfusionTensor
import numpy as np

pred_top_k = np.array([[1, 0, 2],
       [1, 2, 0],
       [2, 0, 1],
       [0, 1, 2],
       [1, 2, 0]])

truth = np.array([1, 2, 0, 2, 0]).reshape(-1,1)

# I know this function works from before
def get_cm_tensor(targets, top_preds, top_k, n_classes, axis=0):
    ordered_cms = np.zeros([top_k, n_classes, n_classes])
    for k in range(top_k):
        unique_idx, unique_idx_counts = np.unique(
            np.stack([targets, top_preds[:, [k]]], axis=axis),
            return_counts=True,
            axis=1,
        )
        np.add.at(
            ordered_cms[k],
            #unique_idx.T,
            tuple(unique_idx.T.reshape(-1, unique_idx.T.shape[-1]).T),
            unique_idx_counts.ravel(),
        )
    return ordered_cms

from scipy.stats import dirichlet
diri = dirichlet([1/3]*3)
samples = diri.rvs(5, random_state=0)

ct = OrderedConfusionTensor(truth, samples, [0,1,2])

top_cms = get_cm_tensor(truth, pred_top_k, 3, 3)

(top_cms == ct.tensor).all()

# Check NDE label encoder when class labels are str
ct_str = OrderedConfusionTensor(
    truth.astype(str),
    samples,
    ['0','1','2'],
    targets_idx=False,
)

(top_cms == ct_str.tensor).all()



from importlib import import_module
__version__ = '0.1.7rc2'

__all__ = [
    'io',
    'data',
    'ml',
    'profile',
    #'ray',
    #'visuals',
]

for module in __all__:
    globals()[module] = import_module(f'.{module}', 'exputils')
del import_module, module



from exputils.data.confusion_matrix import ConfusionMatrix
from exputils.data.ordered_confusion_matrix import OrderedConfusionMatrices
from exputils.data import labels

__all__ = ['ConfusionMatrix', 'OrderedConfusionMatrices']



"""Useful label management classes and functions."""
from copy import deepcopy
from itertools import islice
import numbers

from bidict import OrderedBidict, MutableBidict
import numpy as np
from sklearn.preprocessing import label_binarize
from sklearn.utils.validation import _num_samples as n_samples

import logging
logger = logging.getLogger(__name__)


# TODO add Torch version of this: support for doing this in Torch code only
# TODO add Tensorflow version of this: support for doing this in TF code only
# TODO add JAX version of this: support for doing this in JAX code only
# TODO If it makes sense to, break dep on sklearn


def load_label_set(filepath, sep=None, *args, **kwargs):
    """Loads the given file and reads the labels in. Expects label per line.

    Parameters
    ----------
    filepath : str
        The filepath to the file containing the labels
    sep : str, optional
        The sep character if the file contains provided encodings per
        label. Always assumes one label per line. Will assume first column is
        the original label to be encoded to the provided encoding when
        sep is not None.
    """
    if sep is None:
        with open(filepath, 'r') as openf:
            nd_enc = NominalDataEncoder(
                openf.read().splitlines(),
                *args,
                **kwargs,
            )
        return nd_enc

    # TODO load as csv or tsv, or YAML
    raise NotImplementedError(
        f'sep is not None, but loading CSV not implemented. {sep}',
    )


class BidirDict(dict):
    """A bidirectional dictionary that streamlines the maintenance of two
    dictionaries used together and allows for keys to map to multiple of the
    same values.

    Notes
    -----
        This is modified from the original code provided at:
            https://stackoverflow.com/a/21894086/6557057
    """
    def __init__(self, *args, **kwargs):
        super(BidirDict, self).__init__(*args, **kwargs)
        self.inverse = {}
        for key, value in self.items():
            self.inverse.setdefault(value,[]).append(key)

    def __setitem__(self, key, value):
        if key in self:
            self.inverse[self[key]].remove(key)
        super(BidirDict, self).__setitem__(key, value)
        self.inverse.setdefault(value,[]).append(key)

    def __delitem__(self, key):
        self.inverse.setdefault(self[key],[]).remove(key)
        if self[key] in self.inverse and not self.inverse[self[key]]:
            del self.inverse[self[key]]
        super(BidirDict, self).__delitem__(key)


# TODO perhaps make a NominalDataBidict and have the encoders be numeric
# encodings only? The Bidict would allow for ease for changing labels to others
# but perhaps only bidict and BidirDict are necessary for that, rather than its
# own class. That depends on what else is desired by the NominalDataBidict

# TODO NominalDataMap(): to allow for mapping any nominal data to other nominal
# labels, while the NominalDataEncoder is for encoding nominal data to integers
#   The point for this is to extend OrderedBidict so it already has the ability
#   to to transform numpy arrays when given.


"""
class KeySortedBidict(MutableBidict):
    __slots__ = ()
    _fwdm_cls = SortedDict
    _invm_cls = ValueSortedDict
    _repr_delegate = list

    # TODO the KeySortedBiDict's SortedDict and ValueSortedDict are not
    # Reversible; unable to obtain reverse iterator via reversed(). This needs
    # fixed. For now use OrderedBidict
#"""


class NominalDataEncoder(object):
    """A single class for handling the encoding of nominal data to integer
    values or one hot / binarized vectors.

    nominal_value -> int -> one_hot_vector or binary vector

    Consider extending bidict.

    Attributes
    ----------
    encoder : OrderedBidict
        The bidirectional mapping of nominal value to integer encoding. There
        can be no multiple keyes that map to the same values.
    _argsorted_keys : list = None
        np.ndarray(int) = None
        When the keys in the encoder are not sorted, but instead saved in the
        order they are given, then _argsorted_keys is an array of the indices
        of the encoder keys in sorted order. This is necessary for encoding
        using numpy only when the keys are not sorted when saved into the
        encoder. If the keys are sorted when the encoder is created, then this
        is None.
    pos_label : int = 1
        The positive label to use when binarizing or one hot encoding.
    neg_label : int = 0
        The negative label to use when binarizing or one hot encoding.
    sparse_output : bool = False
        ??? same as scikit LabelBinarizer learn atm.
    unknown_key : str = None
        The key used for an unknown label, especially in encoding of unknown
        or other labels not contained within the encoder. When this is None,
        the default, no handling of unknowns is supported by this encoder.

        This class is different from the others in that nominal values
        encountered in encoding that are not in the set of known labels will be
        treated as this unknown label.

    Notes
    -----
        This is to provide ease-of-use for handling nominal data encodings
        where sklearn's label encoders were not as ergonomic. Here, "ergonomic"
        is about grouping the necessary parts together for handling labels so
        it is all in one place. This class could be implemented to wrap the
        sklearn's encoders, but Bidict was used instead along with sklearn
        functions.
    """
    def __init__(
        self,
        ordered_keys,
        shift=0,
        pos_label=1,
        neg_label=0,
        sparse_output=False,
        ignore_dups=False,
        sort_keys=False,
        unknown_key=None,
        unknown_idx=None,
    ):
        """
        Parameters
        ----------
        ordered_keys : str
            any Iterable
            The keys to be added to the encoder.
        shift : int = 0
            Shifts the encoding by the given value. Can be seen as the starting
            value of the ordered encodings.
        pos_label : see self
        neg_label : see self
            The negative label to use when binarizing or one hot encoding.
        sparse_output : see self
        ignore_dups : bool = False
            Ignores any duplicates in the given ordered keys. Not implemented!
        sort_keys : bool = False
        unknown_key : see self
        unknown_idx : int = None
            The index encoding for the unknown catch-all class, which is only
            used when `unknown_key` is not None. When not given, this defaults
            to shift, which defaults to 0.

            If storing the unknown key within the encoder, we recommend having
            unknown_idx = 0 when used to be represented as some other or
            unknown class that is not captured by the rest of the classes,
            especially when new known classes may be added to the encoder, as
            this keeps unknown consistently at the same indedx, regardless of
            prior or future known classes.
        """
        if not ignore_dups and len(set(ordered_keys)) != len(ordered_keys):
            raise ValueError('There are duplicates in the given sequence')

        if ignore_dups:
            raise NotImplementedError('Ignore_dups is not yet implemented.')

        if neg_label >= pos_label:
            raise ValueError(
                'neg_label ({neg_label}) >= pos_label ({pos_label}), must be <'
            )
        self.pos_label = pos_label
        self.neg_label = neg_label
        self.sparse_output = sparse_output

        if sort_keys:
            # np.unique sorts the keys lexically
            ordered_keys = np.unique(ordered_keys)
        self._argsorted_keys = None
        self._argsorted_unknown_idx = None

        """TODO
        # Sort the keys so they are in the encoder sorted, rather than
        # order given.

        # KeySortedBidict keeps the keys sorted when updated.
        self.encoder = KeySortedBidict({
            key: enc + shift for enc, key in enumerate(ordered_keys, shift)
        })
        #"""
        #else:
        # Use in order given, but maintain sorted_args for encoding
        # TODO probably can make ignore dups and error raise more efficient
        # if already using unique like this. NOTE that unique here does not
        # work to get _argsorted_keys unless ordered_keys is already unique
        # keys only, which is it given the check and NOT ignore_dups

        self.encoder = OrderedBidict({
            key: enc for enc, key in enumerate(ordered_keys, shift)
        })

        # Handle all unknown key internal state and checks
        self._unknown_key = unknown_key
        if unknown_key is not None:
            """
            raise NotImplementedError(' '.join([
                'init setup, but encode needs checked for non object dtypes!',
                'The python encode object dtype works cuz of dict get default',
                'If ints or floats, the unknowns will not default to unknown',
            ]))
            #"""
            if unknown_key in self.encoder:
                if unknown_idx is not None:
                    # Check if the unknown index is correct
                    if unknown_idx != self.encoder[self.unknown_key]:
                        raise ValueError(' '.join([
                            '`unknown_key` in ordered_keys but `unknown idx`',
                            'given a different index than expected:',
                            f'`unknown_idx` is {unknown_idx}, but expected',
                            f'{self.unknown_key} based on ordered keys.',
                        ]))
            elif unknown_idx in {None, shift}:
                self.shift_encoding(1)
                tmp = OrderedBidict({unknown_key: shift})
                tmp.update(self.encoder)
                self.encoder = tmp
            elif unknown_idx in {-1, len(self.encoder) + shift}:
                self.append(unknown_key)
                # TODO unit test to ensure append handles unknown correctly!
            elif (
                isinstance(unknown_idx, int)
                and unknown_idx >= shift
                and unknown_idx <= shift + len(self.encoder)
            ):
                first_half = OrderedBidict(
                    islice(self.encoder.items(), unknown_idx - shift)
                )
                first_half.update(OrderedBidict({unknown_key: unknown_idx}))
                first_half.update(OrderedBidict({
                    k : v + 1 for k, v in
                    islice(self.encoder.items(), unknown_idx - shift, None)
                }))
                self.encoder = first_half

            else:
                raise TypeError(' '.join([
                    '`unknown_idx` is not None or an int within label indices',
                    f'[{shift}, {shift + len(self.encoder)}]!',
                ]))

            #if self._argsorted_keys is not None:

        if not sort_keys:
            # Must update the _argsorted_keys for approriate encoding
            # TODO replace this hotfix cuz this is inefficient!
            #self._argsorted_keys = np.argsort(self.encoder)
            self._update_argsorted_keys(self.encoder)

    def __eq__(self, other):
        if not isinstance(other, NominalDataEncoder):
            logger.warning(
                'Unsupported comparison of NominalDataEncoder object to an '
                'object of type `%s`. Performing surface level duck typing.',
                type(other),
            )
        for key in self.__dict__:
            self_val = getattr(self, key)
            other_val = getattr(other, key)
            try:
                if self_val != other_val:
                    return False
            except ValueError:
                if (self_val != other_val).any():
                    return False
        return (
            self.unknown_key == other.unknown_key
            and self.unknown_idx == other.unknown_idx
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __contains__(self, obj):
        return obj in self.encoder

    def __getitem__(self, key):
        return self.encoder[key]

    def __iter__(self):
        return iter(self.encoder)

    def __reversed__(self):
        return reversed(self.encoder)

    def __len__(self):
        return len(self.encoder)

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for key, val in self.__dict__.items():
            setattr(result, key, deepcopy(val, memo))
        return result

    def _update_argsorted_keys(self, keys):
        """Updates argsorgted keys and the argsorted idx of the first key."""
        unique, self._argsorted_keys = np.unique(keys, return_index=True)
        if self.unknown_key is not None:
            self._argsorted_unknown_idx = np.flatnonzero(
                unique == self.unknown_key
            )[0]

    @property
    def inv(self):
        return self.encoder.inverse

    @property
    def inverse(self):
        return self.encoder.inverse

    @property
    def are_keys_sorted(self):
        """Returns Boolean of if the keys are sorted within this encoder."""
        return self._argsorted_keys is None

    @property
    def unknown_key(self):
        return self._unknown_key

    @property
    def unknown_idx(self):
        return self.encoder.get(self._unknown_key, None)

    @property
    def shift(self):
        return next(iter(self.encoder.inverse))

    def keys(self, *args, **kwargs):
        return self.encoder.keys(*args, **kwargs)

    def values(self, *args, **kwargs):
        return self.encoder.values(*args, **kwargs)

    def items(self, *args, **kwargs):
        return self.encoder.items(*args, **kwargs)

    def get(self, key, default=None):
        return self.encoder.get(key, default)

    def encode(self, keys, one_hot=False):
        """Encodes the given values into their respective encodings.

        Parameters
        ----------
        keys : scalar or np.ndarray
        one_hot : bool = False
            If True, then expects to encode the keys into their respective one
            hot vectors. Otherwise, expects to map elements to their respective
            encoding values.

        Returns
        -------
        scalar or np.ndarray
            Same shape as input keys, but with elements changed to the proper
            encoding.

        Notes
        -----
        When unknown_idx is not None, currently uses numpy.where() after
        checking all np.searchsorted values of 0 and len(self.encoder) and
        checks all values in keys for where the np.seachsorted 0 is an unknown
        label. This is probably inefficient and could be improved.

        TODO, may be beneficial to include a live updating behavior that when
        an unknown token is encountered, rather than throwing an error if
        unknown_key does not exist, or treating that token as the unknown key,
        update the encoder to include this new token. Could be a bool param
        `update` that defaults to False here.
        """
        if one_hot:
            try:
                return label_binarize(
                    keys,
                    classes=np.array(self.encoder),
                    pos_label=self.pos_label,
                    neg_label=self.neg_label,
                    sparse_output=self.sparse_output,
                )
            except ValueError:
                pass

        keys = np.array(keys)

        if n_samples(keys) == 0:
            return np.array([])

        # Check for unrecognized keys # TODO may be able to be more efficient?
        diff = set(np.unique(keys)) - set(self.encoder)
        if diff:
            #   unknowns=None
            if not self._unknown_key:
                raise ValueError(
                    f'`keys` contains previously unseen keys: {diff}',
                )
            #else:
            #    raise NotImplementedError(' '.join([
            #        f'`keys` contains previously unseen keys: {diff}',
            #        'and unknown_key is provided, but not implremented yet!',
            #    ]))

            # TODO XOR allow for updating of the labels in order of occurrence.
            # XOR default is as is, fail if unseen label in encoding.
            #   unknowns = update
            # if update: ...

        shift = self.shift

        if keys.dtype == object:
            # Python encode
            encoded =np.array([
                self.encoder.get(key, self.unknown_idx) for key in keys
            ]).reshape(keys.shape)

        # Numpy encode
        elif self.are_keys_sorted:
            # Encoder keys are already sorted within the encoder.
            encoded = np.searchsorted(self.encoder, keys)
            if self.unknown_idx is not None:
                # Checks where encoded == len(encoder) or (0 and is unknown)
                encoded = np.where(
                    (
                        (encoded != len(self.encoder))
                        & ~ (
                            (encoded == 0)
                            & (
                                keys != (
                                    self.inv[shift + 1]
                                    if self.unknown_idx == shift
                                    else self.inv[shift]
                                )
                            )
                        )
                    ),
                    encoded,
                    self.unknown_idx,
                )
        else:
            # Encoder keys are not necessarily sorted w/in encoder.
            encoded = np.searchsorted(
                self.encoder,
                keys,
                sorter=self._argsorted_keys,
            )
            if self.unknown_idx is None:
                encoded = self._argsorted_keys[encoded]
            else:
                # Checks where encoded == len(encoder) or (0 and is unknown)
                encoded = self._argsorted_keys[np.where(
                    (
                        (encoded != len(self.encoder))
                        & ~ (
                            (encoded == 0)
                            & (
                                keys != (
                                    self.inv[shift + 1]
                                    if self.unknown_idx == shift
                                    else self.inv[shift]
                                )
                            )
                        )
                    ),
                    encoded,
                    self._argsorted_unknown_idx,
                )]

        if one_hot:
            # TODO beware shift for when a one hot encoding!
            # Supporting multioutput data: make new axis last by default
            if self.pos_label != 1 or self.neg_label != 0:
                n_classes = len(self.encoder)
                one_hot_classes = np.empty([n_classes] * 2)
                one_hot_classes.fill(self.neg_label)
                one_hot_classes[np.eye(n_classes, dtype=bool)] = \
                    np.array([self.pos_label] * n_classes)
            else:
                one_hot_classes = np.eye(len(self.encoder))
            return one_hot_classes[encoded]
            # TODO support changing where the added dimension goes.
        elif keys.dtype != object and self.shift != 0:
            logging.debug('shift = %d', self.shift)
            return encoded + self.shift
        return encoded

        # TODO to get this to work w/ np.searchsorted as sklearn does it, a
        # sorted args of the keys must always be present. This means as the
        # keys change, this sorted args must also change. Otherwise, this needs
        # done a different way. This is the cost of having any order of keys.

        #return keys

    def decode(self, encodings, one_hot_axis=None):
        """Decodes the given encodings into their respective keys.

        Parameters
        ----------
        encodings : scalar or np.ndarray
        one_hot : bool
            If True, then expects to decode one hot vectors into their
            respective keys. Otherwise, expects to map elements to their
            respective keys.

        Returns
        -------
        scalar or np.ndarray
            Same shape as input encodings, but with elements changed to the
            proper encoding.
        """
        if isinstance(one_hot_axis, int):
            # TODO check encodings.shape to expected shape
            encodings = encodings.argmax(axis=one_hot_axis)
            if self.shift != 0:
                encodings += self.shift

        # inverse transform of empty array is empty array
        if n_samples(encodings) == 0:
            return np.array([])

        diff = np.setdiff1d(encodings, np.array(self.inv))
        if len(diff):
            raise ValueError(
                f'encodings contains previously unseen labels: {diff}'
            )
            # TODO hard to handle unknowns in the decoding case, but could do
            # update or default as well, I suppose.

        if self.shift != 0:
            return np.array(self.encoder)[np.array(encodings) - self.shift]
        return np.array(self.encoder)[np.array(encodings)]

    def shift_encoding(self, shift):
        """Increments or decrements all encodings by the given integer.

        Parameters
        ----------
        shift : int
            shifts all encodings by this constant integer.
        """
        if not isinstance(shift, int):
            raise TypeError(' '.join([
                'Expected `adjustment` to be type `int`, not',
                f'`{type(shift)}`',
            ]))

        # NOTE uncertain when shift comes into play outside of maintence or
        # when a enc value that is off from that of array indices applies.

        if shift == 0:
            logger.debug('Shift value given was zero. No shifting done.')
            return

        if shift > 0:
            for key in reversed(self.encoder):
                self.encoder[key] += shift
        else:
            for key in self.encoder:
                self.encoder[key] += shift

    def append(self, keys, ignore_dups=False):
        """Appends the keys to the end of the encoder giving them their
        respective encodings.
        """
        # TODO handle the update to _argsorted_keys, more efficiently
        last_enc = next(reversed(self.encoder.inverse))

        if (
            isinstance(keys, list)
            or isinstance(keys, tuple)
            or isinstance(keys, np.ndarray)
        ):
            # Add the multiple keys to the encoder in order.
            for key in keys:
                if key not in self.encoder:
                    last_enc += 1
                    self.encoder[key] = last_enc

                    if not self.are_keys_sorted:
                        # Must update the _argsorted_keys for approriate
                        # encoding TODO replace this hotfix cuz this is
                        # inefficient! # TODO unit test this!
                        #self._argsorted_keys = np.argsort(self.encoder)
                        self._update_argsorted_keys(self.encoder)
                elif ignore_dups:
                    continue
                else:
                    raise KeyError(
                        f'Given key `{key}` is already in the NominalDecoder!',
                    )
        else:
            # Add individual key
            if keys not in self.encoder:
                self.encoder[keys] = last_enc + 1

                if not self.are_keys_sorted:
                    # Must update the _argsorted_keys for approriate encoding
                    # TODO replace this hotfix cuz this is inefficient!
                    # TODO unit test this!
                    #self._argsorted_keys = np.argsort(self.encoder)
                    self._update_argsorted_keys(self.encoder)

            elif ignore_dups:
                return
            else:
                raise KeyError(
                    f'Given key `{keys}` is already in the NominalDecoder!',
                )

    def reorder(self, keys):
        """Reorder the keys"""
        raise NotImplementedError()

        # TODO reorder by new sequence of keys (equivalent to making a new
        # NDEnc but preserving the shift, if there is any, which now may be a
        # depracted thing anyways, so reorder would be superfulous in this case

        # partial reorder, as in swapping class locations, may still be useful.

    def pop(self, key, encoding=False):
        """Pops the single key and updates the encoding as necessary."""
        # NOTE pop key, but then requires updating the rest of the following
        # keys, while if this was done by a list, it would be handled by
        # shifting the array and index mapping done automatically... but then
        # again, iirc, the index mapping runs into a similar issue wrt to
        # getting the index of the keys.

        if key == self.unknown_key:
            self._unknown_key = None
            self._argsorted_unknown_idx = None

        # TODO handle the update to _argsorted_keys

        # Handle the shift in encoding if there is any.
        #shift = next(iter(self.encoder.inverse))

        # Obtain the last encoding
        last_enc = next(reversed(self.encoder.inverse))
        prior_shift = self.shift

        # Remove the given key, whether it is a key or encoding
        if encoding:
            enc = key
            key = self.encoder.inverse.pop(key)
        else:
            enc = self.encoder.pop(key)

        logging.debug('key = `%s`, enc = `%s`', key, enc)

        if enc != last_enc:
            # Decrement all following keys by one
            for key in list(self.encoder)[enc - prior_shift:]:
                logging.debug(
                    'key = `%s`, encoded key = `%s`',
                    key,
                    self.encoder[key],
                )
                self.encoder[key] -= 1

        if not self.are_keys_sorted:
            # Must remove the key's respective arg from _argsorted_keys
            """ TODO fix
            arg = np.argwhere(np.array(self.encoder) == (
                self.encoder.inverse[key] if encoding else key
            ))[0][0]

            self._argsorted_keys = np.delete(self._argsorted_keys, arg)

            # adjust the rest of the args accordingly
            self._argsorted_keys[np.where(self._argsorted_keys > arg)] -= 1
            #"""
            self._update_argsorted_keys(self.encoder)

        return key if encoding else enc

        # TODO efficiently handle the popping of a sequence of keys and the
        # updating of the encoding.

    # TODO consider an insert_after(), inplace of a append() then reorder()

    def save(self, filepath, sep=None):
        """Saves the labels as an ordered list where the index is implied by
        the order of the labels.
        """
        if sep is None:
            with open(filepath, 'w') as openf:
                openf.write('\n'.join([str(x) for x in self.encoder]))
        else:
            raise NotImplementedError(' '.join([
                'Saving as any file using separators other than newlines',
                'between the labels is not yet supported.',
            ]))

    @staticmethod
    def load(filepath, sep=None, *args, **kwargs):
        """Loads the ordered list from the file. Defaults to expect a text file
        where each line contains a single nominal label.

        Args
        ----
        filepath : str
        sep : str = None
        """
        return load_label_set(filepath, sep, *args, **kwargs)


# TODO SparseNominalDataEncoder()
#   Same thing but encoding integers can jump between values, meaning missing
#   values are expected.... dunno when this is necessary

# TODO consider the following, but it is only necessary if there are holes
# in the encoder, which there should never be. Probs best to leave to user
# or child class.
#max_enc : int
#    The maximum integer encoding in the current encoder
#min_enc : int
#    The minimum integer encoding in the current encoder

# TODO class OrdinalDataEncoder: an ordinal data version



"""Kfold Cross Validation wrappers for optionally inverting the role of the
focus fold.
"""
from sklearn.model_selection import KFold, StratifiedKFold


def kfold_generator(
    kfolds,
    features,
    labels=None,
    shuffle=True,
    stratified=False,
    train_focus_fold=False,
    random_seed=None,
):
    """Convenience function that generates the K different indices.

    Parameters
    ----------
    kfolds : int
        The number (K) of folds to create from the data.
    features : array-like, shape (n_samples,)
        The data to be split with the number of samples being the first dim.
    labels : array-like, shape (n_samples,), optional
       The target variable for supervised learning problems. Stratification is
       done based on the labels. The number of samples is the first dimension
       of the features.
    shuffle : bool, optional (default=True)
        Whether to shuffle the data or not. Default is True to shuffle.
    stratified : bool, optional (default=False)
        Attempts to stratify the data if True. Default is False.
    train_focus_fold : bool, optional (default=False)
        Convenience parameter for swapping the order of the other_folds and
        focus_fold, specifically for when the `focus_fold` is to be used for
        training as in-sample data, instead of being used for the evaluation,
        which is the norm.
    random_state : int | RandomState | None, optional (default=None)
        The random state or seed to be used by `sklearn.model_selection.KFold`.
    """
    # Data index splitting
    if stratified:
        fold_indices = StratifiedKFold(kfolds, shuffle, random_seed).split(
            features,
            labels,
        )
    else:
        fold_indices = KFold(kfolds, shuffle, random_seed).split(
            features,
            labels,
        )

    for other_folds, focus_fold in fold_indices:
        # Set the correct train and test indices
        if train_focus_fold:
            yield focus_fold, other_folds
        else:
            yield other_folds, focus_fold


def get_kfold_idx(
    focus_fold,
    kfolds,
    features,
    labels=None,
    shuffle=True,
    stratified=False,
    train_focus_fold=False,
    random_seed=None,
):
    """Gets a specific set of indices from a setup of kfold cv."""
    for i, (train_idx, test_idx) in enumerate(kfold_generator(
        kfolds,
        features,
        labels,
        shuffle,
        stratified,
        train_focus_fold,
        random_seed,
    )):
        if i + 1 == focus_fold:
            return train_idx, test_idx



from setuptools import setup
import re

def get_property(prop, project):
    """Gets the given property by name in the project's first init file."""
    result = re.search(
        r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop),
        open(project + '/__init__.py').read()
    )
    return result.group(1)

with open('README.md', 'r') as f:
    long_description = f.read()

with open('requirements.txt', 'r') as f:
    install_requires = f.read()

project_name = 'exputils'

setup(
    name=project_name,
    version=get_property('__version__', project_name),
    author='Derek S. Prijatelj',
    author_email='derek.s.prijatelj@gmail.com',
    packages=[
        project_name,
        f'{project_name}.data',
        #f'{project_name}.io',
        f'{project_name}.ml',
        f'{project_name}.ray',
        f'{project_name}.visuals',
    ],
    #scripts
    description=' '.join([
        'Convenient functions that are commonly used for running machine',
        'learning experiments.',
    ]),
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=f'https://github.com/prijatelj/{project_name}',
    install_requires=install_requires,
    python_requires='>=3.6',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



"""The general input and output of experiments, including a commonly used
argparse setup with nested namespaces.
"""
import argparse
from copy import deepcopy
from datetime import datetime
import importlib.util
import json
import os
import sys

# TODO import h5py
from packaging import version
import numpy as np

import logging
logger = logging.getLogger(__name__)


# TODO currently NestedNamespace still requires all args to be uniquely
# identified. There is perhaps a better solution that is akin to subparsers,
# but allows for multiple "subparsers" to be called per script call and to
# allow for the subparsers to have overlapping arg names because they are
# inherently nested.

class NestedNamespace(argparse.Namespace):
    """An extension of the Namespace allowing for nesting of namespaces.

    Notes
    -----
    Modified version of hpaulj's answer at
        https://stackoverflow.com/a/18709860/6557057

    Use by specifying the full `dest` parameter when adding the arg. then pass
    the NestedNamespace as the `namespace` to be used by `parser.parse_args()`.
    """
    def __setattr__(self, name, value):
        if '.' in name:
            group, _, name = name.partition('.')
            namespace = getattr(self, group, NestedNamespace())
            setattr(namespace, name, value)
            self.__dict__[group] = namespace
        else:
            self.__dict__[name] = value

    def __getattr__(self, name):
        if '.' in name:
            group, _, name = name.partition('.')

            try:
                namespace = self.__dict__[group]
            except KeyError:
                raise AttributeError

            return getattr(namespace, name)
        else:
            getattr(super(NestedNamespace, self), name)


class NumpyJSONEncoder(json.JSONEncoder):
    """Encoder that handles common Numpy values, and general objects. This also
    works for JSON serializing NestedNamespaces.
    """
    def default(self, o):
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, np.integer):
            return int(o)
        if isinstance(o, np.floating):
            return float(o)

        # TODO either remove or make some form of check if valid.
        return o.__dict__


def filename_append(filename, suffix):
    """Appends text to the end of the filename before the file extention."""
    root, ext = os.path.splitext(filename)

    if ext:
        # TODO add optional to ignore ext if fits some pattern.
        # Perhaps add given extenstion to ease respecting the extention in
        # cases where the existing setup fails

        return f'{root}{suffix}{ext}'

    # If there is no extention, simply append suffix
    return f'{filename}{suffix}'


def create_filepath(
    filepath,
    overwrite=False,
    datetime_fmt='_%Y-%m-%d_%H-%M-%S.%f',
):
    """Ensures the directories along the filepath exists. If the file exists
    and overwrite is False, then the datetime is appeneded to the filename
    while respecting the extention.

    Note
    ----
    If there is no file extension, determined via existence of a period at the
    end of the filepath with no filepath separator in the part of the path that
    follows the period, then the datetime is appeneded to the end of the file
    if it already exists.
    """
    # Check if file already exists
    if not overwrite and os.path.isfile(filepath):
        logger.warning(
            ' '.join([
                '`overwrite` is False to prevent overwriting existing files',
                'and there is an existing file at the given filepath: `%s`',
            ]),
            filepath,
        )

        # NOTE beware possibility of program writing the same file in parallel
        filepath = filename_append(
            filepath,
            datetime.now().strftime(datetime_fmt),
        )

        logger.warning('The filepath has been changed to: %s', filepath)
    else:
        # Ensure the directory exists
        dir_path = os.path.dirname(filepath)
        if dir_path:
            os.makedirs(dir_path, exist_ok=True)

    return filepath


def create_dirs(
    filepath,
    overwrite=False,
    datetime_fmt='_%Y-%m-%d_%H-%M-%S.%f',
):
    """Ensures the directory path exists. Creates a sub folder with current
    datetime if it exists and overwrite is False.
    """
    # Check if dir already exists
    if not overwrite and os.path.isdir(filepath):
        logger.warning(
            ' '.join([
                '`overwrite` is False to prevent overwriting existing',
                'directories and there is an existing file at the given',
                'filepath: %s',
            ]),
            filepath,
        )

        # NOTE beware possibility of a parallel program writing same filename
        filepath = os.path.join(
            filepath,
            datetime.now().strftime(datetime_fmt),
        )
        os.makedirs(filepath, exist_ok=True)

        logger.warning('The filepath has been changed to: %s', filepath)
    else:
        os.makedirs(filepath, exist_ok=True)

    return filepath


def write(file, mode='x', *args, **kwargs):
    """Convenience wrapper for python open file and safely write using
    create_filepath, appending datetime if already existing. This simply
    handles the exclusive write fail case.
    """
    if 'x' not in mode or 'r' in mode or 'w' in mode or '+' in mode:
        raise ValueError(
            "exputils.io.write() only supports exclusive creation. `mode='x'`",
        )

    # Using create_filepath ensures any directories are made and no overwriting
    return open(create_filepath(file), mode, *args, **kwargs)


def save_json(
    filepath,
    results,
    deep_copy=True,
    overwrite=False,
    datetime_fmt='_%Y-%m-%d_%H-%M-%S.%f',
):
    """Saves the content in results and additional info to a JSON.

    Parameters
    ----------
    filepath : str
        The filepath to the resulting JSON.
    results : dict
        The dictionary to be saved to file as a JSON.
    deep_copy : bool
        Deep copies the dictionary prior to saving due to making the contents
        JSON serializable.
    overwrite :
        If file already exists and False, appends datetime to filename,
        otherwise that file is overwritten.
    datetime_fmt : str
        Format of the datetime string that is appeneded to a file whose given
        filepath already exists and overwrite is False.
    """
    if deep_copy:
        results = deepcopy(results)

    # Check if file already exists
    if not overwrite and os.path.isfile(filepath):
        logger.warning(
            ' '.join([
                '`overwrite` is False to prevent overwriting existing files',
                'and there is an existing file at the given filepath: `%s`',
            ]),
            filepath,
        )

        # NOTE beware possibility of a program writing the same file in parallel
        parts = filepath.rpartition('.')
        filepath = parts[0] + datetime.now().strftime(datetime_fmt) \
            + parts[1] + parts[2]

        logger.warning(f'The filepath has been changed to: {filepath}')

    # Ensure the directory exists
    dir_path = filepath.rpartition(os.path.sep)
    if dir_path[0]:
        os.makedirs(dir_path[0], exist_ok=True)

    with open(filepath, 'w') as summary_file:
        json.dump(
            results,
            summary_file,
            indent=4,
            sort_keys=True,
            cls=NumpyJSONEncoder,
        )


# TODO save hd5f


def multi_typed_arg(*types):
    """Returns a callable to check if a variable is any of the types given."""
    # TODO needs unit tested
    def multi_type_conversion(value):
        for arg_type in types:
            try:
                return arg_type(value)
            except TypeError as error:
                print(f'\n{error}\n')
            except ValueError as error:
                print(f'\n{error}\n{type(error)}\n')
        raise argparse.ArgumentTypeError(' '.join([
            f'Arg of {type(value)} is not convertable to any of the types:',
            f'{types}',
        ]))
    return multi_type_conversion


def check_argv(value, arg, optional_arg=True):
    """Checks if the arg was given and checks if its value is one in the given
    iterable. If true to both, then the arg in question is required. This is
    often used to check if another arg is required dependent upon the value of
    another argument, however if the arg in question of being required has a
    default value, then setting it to required is unnecessary.

    Parameters
    ----------
    value : list() | type | object, optional
        Value is expected to be a list of values to check as the value of the
        given arg. If given vlaue is not iterable, then it is treated as a
        single value to be checked. If a type is given, then its type is
        checked.
    arg : str, optional
        The name of the argument whose value is being checked.
    optional_arg : bool, optional
        Flag indicating if the arg being checked is optional or required by
        default. If the arg is optional and is lacking the initial '--', then
        that is added before checking if it exists in sys.argv. Defaults to
        True.
    """
    if optional_arg and arg[:2] != '--':
        arg = '--' + arg

    if arg in sys.argv:
        idx = sys.argv.index(arg)
        if isinstance(value, type):
            print('\n')
            print(f'value ({value}) is of type {type(value)}.')
            print(f'{sys.argv[idx + 1]} is of type {type(sys.argv[idx + 1])}')
            print('\n')
            try:
                value(sys.argv[idx + 1])
                return True
            except:
                return False
        if not hasattr(value, '__iter__'):
            return value == sys.argv[idx + 1]
        return any([x == sys.argv[idx + 1] for x in value])
    return False


def add_logging_args(parser, log_level='WARNING'):
    """Adds the logging args to the arg parser."""
    log_args = parser.add_argument_group('logging', 'Python logging arguments')

    log_args.add_argument(
        '--log_level',
        default=log_level,
        help='The log level to be logged.',
        choices=['NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
        dest='logging.level',
    )

    log_args.add_argument(
        '--log_filename',
        default=None,
        help='The log file to be written to.',
        dest='logging.filename',
    )

    log_args.add_argument(
        '--log_filemode',
        default='a',
        choices=['a', 'w', 'w+'],
        help='The filemode for writing to the log file.',
        dest='logging.filemode',
    )

    log_args.add_argument(
        '--log_format',
        default='%(asctime)s; %(levelname)s: %(message)s',
        help='The logging format.',
        dest='logging.format',
    )

    log_args.add_argument(
        '--log_datefmt',
        default=None,
        #default='%Y-%m-%d_%H-%M-%S',
        help='The logging date/time format.',
        dest='logging.datefmt',
    )

    log_args.add_argument(
        '--log_overwrite',
        action='store_true',
        help=' '.join([
            'If file already exists, giving this flag overwrites that log',
            'file if filemode is "w", otherwise the datetime is appended to',
            'the log filename to avoid overwriting the existing log file.',
        ]),
        dest='logging.overwrite',
    )


def set_logging(log_args):
    """Given an argparse.Namespace, initialize the python logging."""
    # Set logging configuration
    numeric_level = getattr(logging, log_args.level.upper(), None)

    if not isinstance(numeric_level, int):
        raise ValueError(f'Invalid log level given: {log_args.level}')

    if log_args.filename is not None:
        overwrite = (
            log_args.overwrite
            or log_args.filemode not in {'w', 'w+'}
        )

        log_args.filename = create_filepath(log_args.filename, overwrite)

        logging.basicConfig(
            filename=log_args.filename,
            filemode=log_args.filemode,
            level=numeric_level,
            format=log_args.format,
            datefmt=log_args.datefmt,
        )

        if log_args.filemode == 'a':
            # Adding some form of line break for ease of searching logs.
            logger.info('Start of new logging session.')
    else:
        logging.basicConfig(
            level=numeric_level,
            format=log_args.format,
            datefmt=log_args.datefmt,
        )


def add_hardware_args(parser, ml_libs=None):
    """Adds the arguments detailing the hardware to be used."""
    # TODO consider packaging as a dict/NestedNamespace
    # TODO consider a boolean or something to indicate when to pass a
    # tensorflow session or to use it as default
    hardware = parser.add_argument_group(
        'hardware',
        'Hardware uage constraints.',
    )

    hardware.add_argument(
        '--cpus',
        default=1,
        type=int,
        help='The number of available CPUs.',
        dest='hardware.cpus',
    )
    hardware.add_argument(
        '--cpu_cores',
        default=1,
        type=int,
        help='The number of available cores per CPUs.',
        dest='hardware.cpu_cores',
    )
    hardware.add_argument(
        '--gpus',
        default=0,
        type=int,
        help='The number of available GPUs. Pass negative value if no CUDA.',
        dest='hardware.gpus',
    )

    # TODO implement general way of setting which GPU is used for TF & torch
    #hardware.add_argument(
    #    '--which_gpu',
    #    default=None,
    #    type=int,
    #    help='The specifer(s) of which GPU(s) to be used.',
    #)

    hardware.add_argument(
        '--ml_libs',
        default=ml_libs,
        nargs='+',
        help='The machine learning libraries to expect and to init hardware.',
        choices=[
            None,
            'infer',
            'jax',
            'keras',
            'tensorflow',
            'tf',
            'pytorch',
            'torch',
        ],
        dest='hardware.ml_libs',
    )



def set_hardware(args):
    """Sets machine learning librarys' default hardware configuration."""
    # Setup list of ml_libs whose hardware is to be configured
    if args.ml_libs is None:
        # Skip default hardware configuration
        return

    if 'infer' in args.ml_libs:
        # Find ml libraries are installed and initialize for all.
        ml_libs = []

        for lib in ['jax', 'tensorflow', 'keras', 'torch']:
            if importlib.util.find_spec(lib) is not None:
                ml_libs.append(lib)
    else:
        ml_libs = args.ml_libs

    logger.info('Setting the default hardware configuration for %s', ml_libs)

    if 'jax' in ml_libs:
        # TODO add option to provide default config, only add when necessary
        logger.error('JAX default hardware config not implemented yet.')

    if 'keras' in ml_libs:
        import keras
        import tensorflow as tf

        logger.info('Keras version being used: %s', keras.__version__)
        logger.info('Tensorflow version being used: %s', tf.__version__)

        if version.parse(tf.__version__) < version.parse('2.0.0'):
            # Tensorflow 1.15 backend
            keras.backend.set_session(tf.Session(config=get_tf_config(
                args.cpu_cores,
                args.cpus,
                args.gpus,
            )))
        else:
            # TODO handle tf 2.+ config of hardware. Add option to provide
            # default config, only add when necessary
            logger.error('Tensorflow 2.+ default config not implemented yet.')

        # NOTE in the future, may have to add support for other Keras backends
    elif 'tensorflow' in ml_libs or 'tf' in ml_libs:
        import tensorflow as tf

        logger.info('Tensorflow version being used: %s', tf.__version__)

        if version.parse(tf.__version__) < version.parse('2.0.0'):
            # TODO handle tf 1.15 config of hardware. Add option to provide
            # default config, only add when necessary
            logger.error(
                'Tensorflow 1.15 default config not implemented w/o Keras.',
            )
        else:
            # TODO handle tf 2.+ config of hardware. Add option to provide
            # default config, only add when necessary
            logger.error('Tensorflow 2.+ default config not implemented yet.')

    if 'torch' in ml_libs or 'pytorch' in ml_libs:
        # TODO add option to provide default config, only add when necessary
        logger.error('PyTorch default config not implemented yet.')


def get_tf_config(cpu_cores=1, cpus=1, gpus=0, allow_soft_placement=True):
    """Convenient creation of Tensorflow < 2.0 ConfigProto."""
    # TODO Currently only tf 1.15, needs to be updated for 2.+
    from tensorflow import ConfigProto

    return ConfigProto(
        intra_op_parallelism_threads=cpu_cores,
        inter_op_parallelism_threads=cpu_cores,
        allow_soft_placement=allow_soft_placement,
        device_count={
            'CPU': cpus,
            'GPU': gpus,
        } if gpus >= 0 else {'CPU': cpus},
    )


def add_kfold_cv_args(parser):
    """Adds args to the arg parser specific to K-fold cross validation."""

    # TODO simplify this to the kfold params provided in exputils

    kfold_cv = parser.add_argument_group(
        'kfold_cv',
        'K fold cross validation args for evaluating models.',
    )

    kfold_cv.add_argument(
        '--kfolds',
        default=2,
        type=int,
        help='The number of K folds to use in K fold cross validation.',
        dest='kfold_cv.kfolds',
    )

    kfold_cv.add_argument(
        '--no_shuffle_data',
        action='store_false',
        help='Add flag to disable shuffling of data.',
        dest='kfold_cv.shuffle',
    )

    kfold_cv.add_argument(
        '--stratified',
        action='store_true',
        help='Stratified K fold cross validaiton will be used.',
        dest='kfold_cv.stratified',
    )

    kfold_cv.add_argument(
        '--train_focus_fold',
        action='store_true',
        help=' '.join([
            'The focus fold in K fold cross validaiton will be used for ',
            'training and the rest will be used for testing.',
        ]),
        dest='kfold_cv.train_focus_fold',
    )

    # NOTE may remove this from arg group.
    kfold_cv.add_argument(
        '--focus_fold',
        default=None,
        type=int,
        help=' '.join([
            'The focus fold to split the data on to form train and test sets',
            'for a singlemodel train and evaluate session (No K-fold Cross ',
            'Validation; Just evaluates one partition).',
        ]),
        dest='kfold_cv.focus_fold',
    )


def parse_args(
    arg_set=None,
    custom_args=None,
    description=None,
    log_level='WARNING',
    ml_libs=None,
):
    """Creates the args to be parsed and the handling for each.

    Parameters
    ----------
    arg_set : iterable, optional
        contains the additional argument types to be parsed.
    custom_args : function | callable, optional
        Given a function that expects a single argument to be
        `argparse.ArgumentParser`, this function adds arguments to the parser.
    description : str, optional
        The argparse.ArgumentPartser description.

    Returns
    -------
    argparse.namespace
        Parsed arguments.
    """
    parser = argparse.ArgumentParser(description=description)

    # Setup and configure parser
    add_logging_args(parser, log_level=log_level)
    add_hardware_args(parser, ml_libs=ml_libs)

    if arg_set and 'kfold' in arg_set:
        add_kfold_cv_args(parser)

    # TODO add optional profiling args function
    #if arg_set and 'profiling' in arg_set:
    #    add_profiling_args(parser)

    # Add custom args
    if custom_args and callable(custom_args):
        custom_args(parser)

    # Parse args
    args = parser.parse_args(namespace=NestedNamespace())

    # Perform necessary default args setup
    set_logging(args.logging)
    set_hardware(args.hardware)

    # TODO handle random seeds whenever it becomes a necessary thing in scripts
    # e.g. deterministic results instead of statistical reproduction.
    # LOAD the randomseeds from file if it is of type str!!!
    #if isinstance(args.random_seeds, list) and len(args.random_seeds) <= 1:
    #    if os.path.isfile(args.random_seeds[0]):
    #        raise NotImplementedError('Load the random seeds from file.')
    #    else:
    #        args.random_seeds[0] = int(args.random_seeds[0])
    #elif isinstance(args.random_seeds, list):
    #    args.random_seeds = [int(r) for r in args.random_seeds]

    return args



from exputils.io import argconfig

__all__ = [
    'create_dirs',
    'create_filepath',
    'filename_append',
    'NestedNamespace',
    'NumpyJSONEncoder',
    'save_json',
    'write',
]

from exputils.io import __all__

__all__ = ['argconfig'] + __all__
